from base import BasePlan, RollupPlan, BadPlanSituation
from skylark.utils.jsmin import jsmin
from skylark import time_started


class FewestFiles(BasePlan, RollupPlan):
    make_css_urls_absolute = True

    def __split_static_all(self, attr, page_instructions):
        rollup = []
        keep = []
        insert_point = 0

        for item in getattr(page_instructions, attr):
            if 'ieversion' not in item and \
               item.get('include', True) and \
               'static' in item:
                rollup.append(item)
                if not insert_point:
                    insert_point = len(keep)
            else:
                keep.append(item)

        return (rollup, keep, insert_point)

    def prepare_js(self, page_instructions):
        if not hasattr(page_instructions, 'js'):
            return

        self._prepare_assets(page_instructions, ('media/js/templates',))

        rollup, keep, insert_point = self.__split_static_all(
            'js', page_instructions)

        # If there is anything Chirp related, we need to do that here to make it
        # part of our rollup
        super(FewestFiles, self).prepare_chirp(page_instructions)

        rollup.extend(self._rollup_chirp(self.prepared_instructions['chirp']))

        # If there is anything left in the chirp instructions, this plan does
        # not support that situation
        pi_chirp = self.prepared_instructions['chirp']
        if [i['require'] for i in pi_chirp if i['require']]:
            raise BadPlanSituation('Tried to rollup the Javascript into one '
                'file, but there are still separate files from a "chirp:" '
                'instruction.')

        minifier = jsmin if self.options['minify_javascript'] else None

        setattr(page_instructions, 'js', keep)
        self._prepare_file('js', page_instructions)

        self._prepare_rollup('js', rollup, keep, insert_point,
            minifier=minifier)

    def prepare_css(self, page_instructions):
        if not hasattr(page_instructions, 'css'):
            return

        self._prepare_assets(page_instructions, ('media/img',))

        rollup, keep, insert_point = self.__split_static_all(
            'css', page_instructions)

        setattr(page_instructions, 'css', keep)
        self._prepare_file('css', page_instructions)

        self._prepare_rollup('css', rollup, keep, insert_point)

    def prepare_chirp(self, page_instructions):
        # We are going to let the prepare_js handle chirp in this case since we
        # want it to rollup everything JS together.  So there is nothing here
        # to do
        pass



from django.utils.functional import memoize

from base import BasePlan
from fewest import FewestFiles
from reusable import ReusableFiles
from separate import SeparateEverything

__all__ = ['MissingMediaPlan', 'get_plan', 'plan_options', 'get_for_context']
__plan_cache = {}


def get_plan(module, attr):
    plans = __import__(module, globals(), locals(), attr)

    return getattr(plans, attr)
get_plan = memoize(get_plan, __plan_cache, 2)


def plan_options(**kwargs):
    BasePlan.set_options(**kwargs)


class MissingMediaPlan(Exception):
    pass


def get_for_context(context, render_full_page):
    try:
        from skylark.conf import settings
        plan = get_plan(
            settings.SKYLARK_PLANS,
            settings.SKYLARK_PLANS_DEFAULT)

        from base import BasePlan
        if isinstance(plan, BasePlan):
            """
            We actually want the class here, not an instance of it
            """
            raise ValueError('Do not set the plan to an instance of %s, pass '
                 'us the actual class object itself' % plan)

        if not render_full_page:
            # Only supported plan for snippets is separating everything
            return SeparateEverything(context, render_full_page)
        else:
            return plan(context, render_full_page)
    except ImportError:
        if settings.DEBUG:
            return SeparateEverything(context, render_full_page)
        else:
            raise MissingMediaPlan('Could not import the media plan for '
                'Django Skylark media deployment: %s name %s' % (
                    settings.SKYLARK_PLANS,
                    settings.SKYLARK_PLANS_DEFAULT,
                ))



import copy
import yaml

from django import template


class StringWithSourcefile(object):
    """
    A string that include the source file from whence it came
    """
    def __init__(self, s, sourcefile):
        self.value = s
        self.sourcefile = sourcefile

    def __str__(self):
        return self.value.__str__()

    def __call__(self, f):
        return self.value.__call__(f)


class PageInstructions(object):
    """
    Object to contain the page instructions that come from our YAML files
    """
    def __init__(self, **kwargs):
        self.render_full_page = kwargs.get('render_full_page', True)
        self.context = kwargs.get('context_instance', template.Context())

        """
        These are the ones we expect, these are all the possible kinds of
        instructions that can be handled
        """
        self.root_yaml = None
        self.doctype = None
        self.body = None
        self.title = None
        self.uses_yaml = []
        self.other_yaml = []
        self.piped_yaml = []
        self.js = []
        self.css = []
        self.meta = []
        self.chirp = []

    def part_exists(self, part):
        """
        Withing our existing page instruction for javascript and css, we look
        to see if the part already has been added.  There is no reason to
        duplicate either one of these.
        """
        instructions = []
        instructions.extend(self.js)
        instructions.extend(self.css)
        instructions.extend(self.chirp)

        for instruction in instructions:
            if self._get_source_attribute(instruction) == \
               self._get_source_attribute(part):
                return True

        return False

    def _get_source_attribute(self, part):
        """
        Looks through a part of the yaml file and pulls out the location it
        references.

        Since the parts can have a variety of keys (url, render, static,
        namespace) this method is used to get whichever of the keys is set
        """
        for attr in ('url', 'inline', 'static', 'namespace'):
            if attr in part:
                return part[attr]

        return None

    def pipe_media_to(self, destination_instructions):
        """
        Push all the media into the destination_instructions
        """
        for attr in ('js', 'css', 'chirp',):
            media = getattr(self, attr)
            to_remove = []
            dest_media = getattr(destination_instructions, attr)

            if not media:
                continue

            for part in media:
                # We are piping all the media, so remove it from this one
                to_remove.append(part)
                if destination_instructions.part_exists(part):
                    continue
                if part['sourcefile'] not in destination_instructions.yaml:
                    destination_instructions.piped_yaml.append(
                        part['sourcefile'])
                dest_media.append(part)

            for part in to_remove:
                media.remove(part)

    @property
    def yaml(self):
        if self.root_yaml:
            yield self.root_yaml

        for uses in self.uses_yaml:
            yield uses

        for other in self.other_yaml:
            yield other

        for piped in self.piped_yaml:
            yield piped

    def __get_object(self, yamlfile, context):
        source = template.loader.get_template(yamlfile)
        assert source, 'The template loader found the template but it is ' + \
            'completely empty'

        sourcerendered = source.render(context)
        assert sourcerendered, 'yamlfile needs to contain something'

        return yaml.load(sourcerendered)

    def add(self, instructions, sourcefile, **kwargs):
        if not self.root_yaml:
            # If this is the first time we see this, it's got to be the root
            self.root_yaml = sourcefile

        if not sourcefile == self.root_yaml:
            # This must be coming from someone who is invoking the add method
            # through PageAssembly.add_page_instruction.  We need to put this
            # in the proper spot, so make sure we aren't adding this to the
            # other_yaml dictionary if it's already in the uses_yaml dictionary
            if not sourcefile in self.uses_yaml:
                self.other_yaml.append(sourcefile)

        """
        Adds the instructions from one YAML file to this object, combining
        it with what's already here
        """
        if not isinstance(instructions, list) or not \
           isinstance(instructions, tuple):
            instructions = (instructions, )

        for instruction in instructions:
            if 'uses' in instruction:
                for uses in instruction['uses']:
                    if uses['file'] == self.root_yaml or \
                       uses['file'] in self.other_yaml or \
                       uses['file'] in self.uses_yaml:
                        # We've already seen this guy, there is no reason to
                        # add this YAML again
                        continue
                    self.uses_yaml.append(uses['file'])
                    self.add(
                        self.__get_object(uses['file'], self.context),
                        uses['file'])

            for attr in ('doctype', 'js', 'css', 'body', 'title', 'meta',
                         'chirp'):
                if attr in instruction:
                    pi_object = getattr(self, attr)
                    i_object = instruction[attr]

                    if isinstance(pi_object, list):
                        for part in i_object:
                            if attr in ('js', 'css', 'chirp',):
                                if self.part_exists(part):
                                    continue

                            part['sourcefile'] = sourcefile
                            pi_object.append(part)
                    else:
                        str_plus_source = StringWithSourcefile(
                            i_object, sourcefile)
                        setattr(self, attr, str_plus_source)



from functools import partial
from django import http
from django.conf import settings
from django.template import loader, TemplateDoesNotExist
from skylark import RequestContext
from skylark.page import PageAssembly

def direct_to_yaml(yaml_file, **kwargs):
    if settings.DEBUG:
        # Let's make sure we can grab this
        loader.find_template(yaml_file)

    def render(request, **kwargs):
        response = kwargs.get('response_object')
        yaml_file = kwargs.get('yaml_file')
        context = RequestContext(request)
        pa = PageAssembly(yaml_file, context)
        return response(pa.dumps())

    response_object = kwargs.get('response_object',
        http.HttpResponse)
    func = partial(render, yaml_file=yaml_file,
        response_object=response_object)
    funcname = 'direct_to_yaml%s' % id(func)
    globals()[funcname] = func

    return 'skylark.views.generic.%s' % funcname

def render_404_from_yaml(yaml_file):
    return direct_to_yaml(yaml_file,
        response_object=http.HttpResponseNotFound)

def render_500_from_yaml(yaml_file):
    return direct_to_yaml(yaml_file,
        response_object=http.HttpResponseServerError)



import default as default_settings
from django.conf import settings as django_settings
from django.utils.functional import LazyObject


class SkylarkSettings(LazyObject):
    def _setup(self):
        for name in dir(default_settings):
            if name == name.upper() and \
               not hasattr(django_settings, name):
                value = getattr(default_settings, name)
                setattr(django_settings, name, value)

        self._wrapped = django_settings

settings = SkylarkSettings()



import os
import shutil
from time import time
from django.http import HttpRequest
from django.core.management import setup_environ

import settings as settings_module
setup_environ(settings_module)
from django.conf import settings

projectdir = os.path.join(os.path.dirname(__file__))

cachedir = settings.SKYLARK_CACHE_ROOT


def exist(*filelist):
    for file in filelist:
        assert os.path.isfile(os.path.join(cachedir, file))


def get_one_file_in(path):
    started = time()
    while (time() - started < 3.0):
        for file in os.walk(path):
            files = file[2] # the third element is an array of files
            if files:
                return os.path.join(file[0], files[0])

    raise Exception, 'Could not find a file in %s' % path


def get_contents(filename):
    try:
        f = open(filename, 'r')

        content = f.read()
        f.close()

        return content
    except IOError:
        from pprint import PrettyPrinter
        pformat = PrettyPrinter(indent=4).pformat
        raise Exception('Could not file %s, options are\n%s' % (filename,
            pformat(os.listdir(os.path.dirname(filename)))))


def get_request_fixture():
    request = HttpRequest()
    request.path = '/'
    request.META = { 'REMOTE_ADDR': '127.0.0.1', 'SERVER_NAME': '127.0.0.1', 'SERVER_PORT': '8000' }
    return request


def setup():
    settings.DEBUG = True
    settings.SKYLARK_PLANS = 'mediadeploy'
    settings.SKYLARK_PLANS_DEFAULT = 'default'
    settings.SKYLARK_PLANS_ROLLUP_SALT = 'aaaaaaaaaaaaaaaa'
    settings.SKYLARK_ENABLE_TIDY = False


def teardown():
    # Remove everything but the addons
    skip = ['addon']
    try:
        for topdir in os.listdir(cachedir):
            if topdir in skip: continue
            shutil.rmtree(os.path.join(cachedir, topdir))
    except OSError:
        pass



from skylark.renderer.base import Renderer
from skylark.renderer.html import *
from skylark.renderer.xhtml import *

renderers = {
    'html': Html5,
    'HTML 4.01 Transitional': Html401Transitional,
    'HTML 4.01 Strict': Html401Strict,
    'HTML 4.01 Frameset':Html401Frameset,
    'XHTML 1.0 Transitional': Xhtml1Transitional,
    'XHTML 1.0 Strict': Xhtml1Strict,
    'XHTML 1.0 Frameset': Xhtml1Frameset,
}
    

def get(doctype, instructions, context, **kwargs):
    return renderers[str(doctype)](
        instructions, context, **kwargs)



"""
From the W3 Spec

URI values (Uniform Resource Identifiers, see [RFC3986], which includes URLs,
URNs, etc) in this specification are denoted by <uri>. The functional notation
used to designate URIs in property values is "url()", as in:

    Example(s):

        body { background: url("http://www.example.com/pinkish.png") }

    The format of a URI value is 'url(' followed by optional white space
    followed by an optional single quote (') or double quote (") character
    followed by the URI itself, followed by an optional single quote (') or
    double quote (") character followed by optional white space followed by
    ')'. The two quote characters must be the same.

Example(s):

    An example without quotes:

        li { list-style: url(http://www.example.com/redball.png) disc }

    Some characters appearing in an unquoted URI, such as parentheses, commas,
    white space characters, single quotes (') and double quotes ("), must be
    escaped with a backslash so that the resulting URI value is a URI token:
        '\(', '\)', '\,'.

    Depending on the type of URI, it might also be possible to write the above
    characters as URI-escapes (where "(" = %28, ")" = %29, etc.) as described
    in [RFC3986].

    In order to create modular style sheets that are not dependent on the
    absolute location of a resource, authors may use relative URIs. Relative
    URIs (as defined in [RFC3986]) are resolved to full URIs using a base URI.
    RFC 3986, section 5, defines the normative algorithm for this process. For
    CSS style sheets, the base URI is that of the style sheet, not that of the
    source document. 
"""
import re
from os.path import normpath

#[\"'\(\)]
url_value_re = re.compile(r"""
   url\(
       (\"|')?
        \s*
        ((
            [^\"'\(\)]|(?<=\\)
        )*)
        \s*
        (\"|')?
    \)
      """, re.VERBOSE)


def relative_replace(source, css_path, cache_base_url):
    urls_replaced = []   # List of urls we've already replaced so we can skip
    replacement = source
    for match in url_value_re.finditer(source):
        url_value = match.group()
        if url_value in urls_replaced:
            # We've already replaced this one
            continue
        urls_replaced.append(url_value)
        
        # The url under question
        url = match.group(2)

        if url.startswith('http') or url.startswith('https') or \
           url.startswith('/'):
            # We need to leave this sucker alone, it's not the droids that we
            # are looking for.  It's either an absolute url or 
            # fully-qualified.
            continue

        full_path = normpath('/'.join((css_path, url,)))
        new_url_value = '%s%s' % (cache_base_url, full_path)

        # We are going to borrow os.path.normpath to fix any double slashes or
        # any ../ parts
        replacement = replacement.replace(url_value,
            'url(%s%s%s)' % (match.group(1) or '', new_url_value,
                             match.group(4) or ''))

    return replacement



from base import BasePlan, RollupPlan
from skylark.utils.jsmin import jsmin
from skylark import time_started


class ReusableFiles(BasePlan, RollupPlan):
    make_css_urls_absolute = True

    def __split_static_uses(self, attr, page_instructions):
        rollup = []
        keep = []
        insert_point = 0

        for item in getattr(page_instructions, attr):
            if item['sourcefile'] in page_instructions.uses_yaml and \
               'ieversion' not in item and \
               item.get('include', True) and \
               'static' in item:
                rollup.append(item)
                if not insert_point:
                    insert_point = len(keep)
            else:
                keep.append(item)

        return (rollup, keep, insert_point)

    def prepare_js(self, page_instructions):
        if not hasattr(page_instructions, 'js'):
            return

        self._prepare_assets(page_instructions, ('media/js/templates',))

        rollup, keep, insert_point = self.__split_static_uses(
            'js', page_instructions)

        minifier = jsmin if self.options['minify_javascript'] else None

        setattr(page_instructions, 'js', keep)
        self._prepare_file('js', page_instructions)

        self._prepare_rollup('js', rollup, keep, insert_point,
            minifier=minifier)

    def prepare_css(self, page_instructions):
        if not hasattr(page_instructions, 'css'):
            return

        self._prepare_assets(page_instructions, ('media/img',))

        rollup, keep, insert_point = self.__split_static_uses(
            'css', page_instructions)

        setattr(page_instructions, 'css', keep)
        self._prepare_file('css', page_instructions)

        self._prepare_rollup('css', rollup, keep, insert_point)

    def prepare_chirp(self, page_instructions):
        super(ReusableFiles, self).prepare_chirp(page_instructions)

        rollup = self._rollup_chirp(self.prepared_instructions['chirp'])

        # If there is anything left in the chirp instructions, we need to wrap
        # our rolled up chirp content with a dojo.addOnLoad()
        pi_chirp = self.prepared_instructions['chirp']
        if [i['require'] for i in pi_chirp if i['require']]:
            wrap_source = ('dojo.addOnLoad(function() {', '});',)
        else:
            wrap_source = ('', '',)

        minifier = jsmin if self.options['minify_javascript'] else None

        self._prepare_rollup('js', rollup, page_instructions.js,
             len(self.prepared_instructions['js']), minifier=minifier,
             wrap_source=wrap_source)



import os
from random import sample
import string

from django.conf import settings as django_settings
from urlparse import urljoin


# General settings
if hasattr(django_settings, 'STATIC_ROOT'):
    SKYLARK_CACHE_ROOT = os.path.join(django_settings.STATIC_ROOT, 'cfcache')
    SKYLARK_CACHE_URL = urljoin(django_settings.STATIC_URL, 'cfcache/')

else:
    SKYLARK_CACHE_ROOT = os.path.join(django_settings.MEDIA_ROOT, 'cfcache')
    SKYLARK_CACHE_URL = urljoin(django_settings.MEDIA_URL, 'cfcache/')

SKYLARK_INIT_CLEAR_CACHE = False
SKYLARK_PLANS = 'mediadeploy'
SKYLARK_PLANS_DEFAULT = 'default'
SKYLARK_PLANS_ROLLUP_SALT = ''.join(sample(string.hexdigits, 16))
SKYLARK_ENABLE_TIDY = False   # For now until tidylib catches up with html5
SKYLARK_RAISE_CSS_ERRORS = django_settings.DEBUG
SKYLARK_RAISE_HTML_ERRORS = django_settings.DEBUG

# Dojo toolkit related
SKYLARK_DOJO_DEBUGATALLCOSTS = django_settings.DEBUG
SKYLARK_DOJO_COPY_INTERNALBUILD = True
SKYLARK_DOJO_VIA_INTERNALBUILD = True
SKYLARK_DOJO_VIA_CDN_AOL = None
SKYLARK_DOJO_VIA_URL = None

# Chirp testing and MVC
SKYLARK_CHIRP_INSTRUMENTED = False



from django import template

from skylark.snippet import SnippetAssembly

register = template.Library()


@register.tag
def tt_snippet_assembly(parser, token):
    return TtSnippetAssemblyNode()


class TtSnippetAssemblyNode(template.Node):
    def render(self, context):
        sa = SnippetAssembly("dummyapp/issue_bb_23/tt_sa.yaml", context)
        return sa.dumps();


@register.tag
def tt_after_snippet_assembly(parser, token):
    return TtAfterSnippetAssemblyNode()


class TtAfterSnippetAssemblyNode(template.Node):
    def render(self, context):
        sa = SnippetAssembly("dummyapp/issue_bb_23/tt_after_sa.yaml", context)
        return sa.dumps();



from skylark.renderer import Renderer


class Xhtml1Strict(Renderer):
    doctype = '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">'
    template_name = 'skylark/xhtml1.html'
    snippet_template_name = 'skylark/xhtmlsnippet.html'


class Xhtml1Transitional(Renderer):
    doctype = '<!DOCTYPE HTML PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" "http://www.w3.org/TR/xhtml1/xhtml1-loose.dtd">'
    template_name = 'skylark/xhtml1.html'
    snippet_template_name = 'skylark/xhtmlsnippet.html'


class Xhtml1Frameset(Renderer):
    doctype = '<!DOCTYPE HTML PUBLIC "-//W3C//DTD XHTML 1.0 Frameset//EN" "http://www.w3.org/TR/xhtml1/xhtml1-frameset.dtd">'
    template_name = 'skylark/xhtml1.html'
    snippet_template_name = 'skylark/xhtmlsnippet.html'



import hashlib
import yaml
import re

from django import http, template
from django.core.urlresolvers import resolve

from skylark import HttpResponse, RequestContext
from skylark.conf import settings
from skylark.instructions import PageInstructions
from skylark import renderer
from skylark import chirp
from skylark.chirp import check_instrumentation

try:
    import tidylib

    tidylib.BASE_OPTIONS = {
        "indent": 1,
        "indent-spaces": 4,
        "quote-marks": 1,
        "tidy-mark": 0,
        "wrap": 0,
        "indent-cdata": 1,
        "force-output": 0,
    }
except OSError:
    """
    We can't use tidy, most likely libtidy is not installed on the system
    """
    settings.SKYLARK_ENABLE_TIDY = False


class HtmlTidyErrors(Exception):
    pass


class HandlerError(Exception):
    pass


class BaseAssembly(object):
    """
    The yaml files we were constructed with, this probably came from a Django
    view
    """
    yamlfiles = None

    """
    The Django template context object we will render everything with
    """
    context = None

    """
    Are we going to render as a full page or a snippet?
    """
    render_full_page = True

    """
    A list of handlers that can be used to alter the Page Assembly before it
    renders content
    """
    _page_assembly_handlers = []

    def __init__(self, yamlfiles, context):
        if not hasattr(self, 'render_full_page'):
            raise ValueError('You must set render_full_page to True '
                'or False on the subclass of BaseAssembly')

        if not isinstance(context, RequestContext):
            raise ValueError('%r must be of type skylark.RequestContext, '
                'you provided %s' % (context, type(context),))

        if not yamlfiles:
            raise ValueError('%r argument must not be empty or None' %
                 (yamlfiles,))

        if not len(yamlfiles):
            raise ValueError('%r must not be zero length' % (yamlfiles,))

        # We can get the yamlfiles as a single string or a tuple, but by the
        # time we set our instance variable it needs to be normalized to a
        # tuple.

        # In other words, this needs to happen:
        #     'some/file.yaml'   -->  ('some/file.yaml',)
        if isinstance(yamlfiles, str):
            yamlfiles = (yamlfiles,)

        self.yamlfiles = yamlfiles

        # It's important for our context to have the media root and url
        # included, if it's not there we are going to add it
        for var in ('MEDIA_ROOT', 'MEDIA_URL'):
            if not hasattr(context, var):
                context[var] = getattr(settings, var)

        self.context = context

        # We keep track of the stack of assemblies that we process, so add this
        # one to the stack
        self.context['skylark_internals']['assembly_stack'].append(self)

    @staticmethod
    def register_handler(handler):
        if not callable(handler):
            raise HandlerError(
                'Handler provided cannot be registered, it is not callable')
        if handler in BaseAssembly._page_assembly_handlers:
            return
        BaseAssembly._page_assembly_handlers.append(handler)

    @staticmethod
    def unregister_handler(handler):
        try:
            BaseAssembly._page_assembly_handlers.remove(handler)
        except ValueError:
            pass

    @staticmethod
    def unregister_all():
        del BaseAssembly._page_assembly_handlers[:]

    def __is_root_assembly(self):
        return True if self == self.__get_root_assembly() else False

    def __get_root_assembly(self):
        astack = self.context['skylark_internals']['assembly_stack']
        return astack[0]

    def __create_page_instructions(self):
        """
        Combines all the files and instructions into one object
        """
        page_instructions = PageInstructions(
            render_full_page=self.render_full_page,
            context_instance=self.context)

        tried = []

        for file in self.yamlfiles:
            self.add_page_instructions(page_instructions, file)

        return page_instructions

    def __convert_tidy_errors(self, errors_raw, **kwargs):
        """
        The errors we get back from the tidy are formatted for output including
        newlines in the string.  Each line is a single error, so we split on
        the line to get a list
        """
        filter = kwargs.get('filter', None)
        pattern = '^line\ (?P<l>\d+)\ column\ (?P<c>\d+)\ \-\ (?P<desc>.*)$'
        line_re = re.compile(pattern)

        errors = []
        for e in errors_raw.split('\n'):
            try:
                line, column, desc = line_re.match(e).groups()
                if [True for i in filter if i in desc]:
                    continue
                errors.append({
                    'line': int(line),
                    'column': int(column),
                    'desc': desc,
                })
            except AttributeError:
                pass

        return errors

    def __format_tidy_errors(self, document_raw, errors):
        document = document_raw.split('\n')

        for error in errors:
            yield '%s: (%s; line %s)' % (
                error['desc'],
                document[error['line'] - 1],
                error['line'],
            )

    def add_page_instructions(self, page_instructions, file):
        source = template.loader.get_template(file)
        assert source, 'The template loader found the template but it is ' + \
            'completely empty'

        sourcerendered = source.render(self.context)
        assert sourcerendered, 'yamlfile needs to contain something'

        instructions = yaml.load(sourcerendered)
        page_instructions.add(instructions, file)

    @check_instrumentation
    def dumps(self):
        self.instructions = self.__create_page_instructions()

        if not self.__is_root_assembly():
            # We are not the root assembly, it's a good bet that we are a
            # SnippetAssembly being rendered within
            root_pa = self.__get_root_assembly()

            # Our media needs to be piped to the root page instructions, we
            # don't want to render media with the SnippetAssembly, the
            # PageAssembly should handle this for us
            self.instructions.pipe_media_to(root_pa.instructions)

        doctype = self.instructions.doctype or 'html'

        page_renderer = renderer.get(doctype, self.instructions, self.context,
            render_full_page=self.render_full_page,
            omit_media=not self.__is_root_assembly())

        for handler in BaseAssembly._page_assembly_handlers:
            handler(self.instructions, page_renderer, self)

        content = page_renderer.render()

        if not settings.DEBUG or not self.render_full_page:
            return content

        if not settings.SKYLARK_ENABLE_TIDY:
            document = content
        else:
            document, errors = tidylib.tidy_document(content)
            # We want to let the proprietary attributes slide, Chirp uses these
            errors = self.__convert_tidy_errors(
                errors, filter=['proprietary attribute'])
            if errors and settings.SKYLARK_RAISE_HTML_ERRORS:
                formatted_errors = self.__format_tidy_errors(content, errors)
                raise HtmlTidyErrors('We tried to tidy up the document and '
                   'got these errors: %s' % ', '.join(formatted_errors))

        return unicode(document or content)

    def get_http_response(self):
        """
        Returns an HttpResponse object will all the combined goodness
        """
        return HttpResponse(self.dumps())



import os
import time
import shutil

from django import http, template
from django.conf import UserSettingsHolder

from skylark.conf import settings
from skylark import templatepatch

__all__ = ['clear_media_cache', 'time_started',
           'HttpResponse', 'RequestContext']

__time_started = time.time()


def time_started():
    return __time_started


class HttpResponse(http.HttpResponse):
    """
    Essentially, a copy of Django's version.  We are making our own here in
    anticipation of some additional features in the future.
    """
    def __init__(self, content='', mimetype=None, status=None,
        content_type=None):
        super(HttpResponse, self).__init__(
            content, mimetype, status, content_type)


def copy_then_filter_settings(values, key=None):
    fs = UserSettingsHolder({})

    for name in dir(values):
        if name == name.upper() and \
           (key is None or key in name):
            value = getattr(values, name)
            setattr(fs, name, value)
    return fs


class RequestContext(template.RequestContext):
    """
    Adds the raw request to the object, we need this to perform some caching
    work later on
    """
    def __init__(self, request, dict=None, processors=None):
        from django.conf import settings as django_settings
        super(RequestContext, self).__init__(request, dict, processors)
        if hasattr(request, 'skylark_internals'):
            internals = request.skylark_internals
        else:
            # We want all the RequestContext created from this request to share
            # this dict
            internals = {
                'request': request,
                'assembly_stack': [],
                'settings': copy_then_filter_settings(
                    django_settings, 'SKYLARK')}
            # Hook the request object
            setattr(request, 'skylark_internals', internals)

        self['skylark_internals'] = internals

__cache_cleared = False


def clear_media_cache():
    cachedir = settings.SKYLARK_CACHE_ROOT
    # Some directories in the cache should not be deleted
    skip = ['addon']
    if not os.path.isdir(cachedir):
        return
    for topdir in os.listdir(cachedir):
        if topdir in skip:
            continue
        d = os.path.join(cachedir, topdir)
        if os.path.isdir(d):
            shutil.rmtree(d)

if settings.SKYLARK_INIT_CLEAR_CACHE and not __cache_cleared:
    """
    When we initialize CF, let's delete the cache automatically if our settings
    tell us to
    """
    clear_media_cache()
    __cache_cleared = True

__copy_addons = False


def copy_addons():
    """
    There are dependencies that Django Skylark and Chirp have that we need to
    include whenever the application starts.

    Right now this is:
        * Dojo (slimmed down, custom build from ext/chirp_dojo.profile.js)

    At the moment, this is slime enough that we are recreating these
    directories everytime the app initializes.  At some point this may become
    unnaceptable.  As the things we use from Dojo increases, where does it make
    sense to just use the entire Dojo release?  This may eventually get altered
    so that it downloads a release of Dojo and untars it into addon.
    """
    addondir = os.path.join(settings.SKYLARK_CACHE_ROOT, 'addon')
    if os.path.isdir(addondir):
        # Don't copy it again, that's silly
        return
    thisdir = os.path.dirname(__file__)
    mediadir = ['templates', 'chirp', 'media', ]
    dojodir = os.path.join(thisdir, *(mediadir + ['dojo']))
    shutil.copytree(dojodir, os.path.join(addondir, 'dojo'))
    dojoxdir = os.path.join(thisdir, *(mediadir + ['dojox']))
    shutil.copytree(dojoxdir, os.path.join(addondir, 'dojox'))

if settings.SKYLARK_DOJO_COPY_INTERNALBUILD and not __copy_addons:
    copy_addons()
    __copy_addons = True



from django.conf.urls.defaults import *
from skylark.views.testrunner import *

urlpatterns = patterns('',
    url(r'^$', testrunner, name='chirp-testrunner'),
    url(r'^subject/start$', subject_start, name='chirp-testrunner-subject-start'),
    url(r'^deinstrument$', deinstrument, name='chirp-testrunner-deinstrument'),
)



# Django settings for example project.
import os
from urlparse import urljoin

PROJECT_PATH = os.path.abspath(os.path.dirname(__file__))

DEBUG = True
TEMPLATE_DEBUG = DEBUG

ADMINS = (
    # ('Your Name', 'your_email@domain.com'),
)

MANAGERS = ADMINS

DATABASE_ENGINE = 'sqlite3'
DATABASE_NAME = 'test.db'
DATABASE_USER = ''
DATABASE_PASSWORD = ''
DATABASE_HOST = ''
DATABASE_PORT = ''

# Local time zone for this installation. Choices can be found here:
# http://en.wikipedia.org/wiki/List_of_tz_zones_by_name
# although not all choices may be available on all operating systems.
# If running in a Windows environment this must be set to the same as your
# system time zone.
TIME_ZONE = 'America/Chicago'

# Language code for this installation. All choices can be found here:
# http://www.i18nguy.com/unicode/language-identifiers.html
LANGUAGE_CODE = 'en-us'

SITE_ID = 1

# If you set this to False, Django will make some optimizations so as not
# to load the internationalization machinery.
USE_I18N = True

# Absolute path to the directory that holds media.
# Example: "/home/media/media.lawrence.com/"
MEDIA_ROOT = os.path.join(PROJECT_PATH, 'media')

# URL that handles the media served from MEDIA_ROOT. Make sure to use a
# trailing slash if there is a path component (optional in other cases).
# Examples: "http://media.lawrence.com", "http://example.com/media/"
MEDIA_URL = 'http://localhost:8000/media/'

# URL prefix for admin media -- CSS, JavaScript and images. Make sure to use a
# trailing slash.
# Examples: "http://foo.com/media/", "/media/".
ADMIN_MEDIA_PREFIX = '/media/'

# Make this unique, and don't share it with anybody.
SECRET_KEY = 'j3z952)d8-4m44vz=q(y%4riczf^ju(2f4oy8*zzhbexq-r@=n'

TEMPLATE_LOADERS = (
    'django.template.loaders.filesystem.Loader',
    'django.template.loaders.app_directories.Loader',
)

MIDDLEWARE_CLASSES = (
    'django.middleware.common.CommonMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
)

ROOT_URLCONF = 'urls'

TEMPLATE_DIRS = (
    # Put strings here, like "/home/html/django_templates" or "C:/www/django/templates".
    # Always use forward slashes, even on Windows.
    # Don't forget to use absolute paths, not relative paths.
    os.path.join(PROJECT_PATH, 'templates')
)

INSTALLED_APPS = (
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.sites',
    'skylark',
    'dummyapp',
    'dynamicapp',
    'planapp',
)

SKYLARK_CACHE_ROOT = os.path.join(MEDIA_ROOT, 'cfcache')
SKYLARK_CACHE_URL  = urljoin(MEDIA_URL, 'cfcache/')



from django import template
from django.template.loader import render_to_string

from skylark.snippet import SnippetAssembly, RequestContext

register = template.Library()

@register.tag
def snippet_assembly(parser, token):
    return SnippetAssemblyNode()

class SnippetAssemblyNode(template.Node):
    def render(self, context):
        sa = SnippetAssembly("dummyapp/issue24/sa.yaml", context)
        return sa.dumps();



import planapp
from skylark import plans
from django.conf import settings

se = plans.SeparateEverything
rf = plans.ReusableFiles
ff = plans.FewestFiles

if settings.DEBUG:
    default = se
else:
    #default = plans.FewestFiles
    default = rf

"""
Eventually we want to be able to apply settings for specific views, this seemed
like a good idea when we were designing this feature but it may turn out to be
bunk.
"""
#plans.apply_to(planapp.views.index, plans.FewestFiles)
#default.apply_to(views.full.someview, plans.SeparateEverything)



import py.test
import os
from time import sleep

from nose.tools import with_setup
from nose.plugins.attrib import attr
from nose.plugins.skip import SkipTest
from django.template import TemplateDoesNotExist

from skylark import *
from skylark import chirp
from skylark import plans
from skylark.page import PageAssembly
from skylark.instructions import PageInstructions
from skylark.plans import *
from skylark.plans.base import BadOption
from skylark.plans.separate import SeparateEverything
from skylark.plans.fewest import FewestFiles

from skylark.tests import *


def test_plan_options_invalid():
    py.test.raises(BadOption, plan_options, not_a_valid_option=True)


@with_setup(setup, teardown)
def test_can_change_deploy_plan_name():
    context = {}
    render_full_page = True

    plan = get_for_context(context, render_full_page)
    assert isinstance(plan, SeparateEverything)

    # Make sure that if there is no deploy plan (file missing) that it comes
    # back with SeparateEverything
    settings.SKYLARK_PLANS = 'mediadeploy_notthere'
    plan = get_for_context(context, render_full_page)
    assert isinstance(plan, SeparateEverything)

    # Change it to something valid that is not the default
    settings.SKYLARK_PLANS = 'mediadeploy_alt'
    settings.SKYLARK_PLANS_DEFAULT = 'alternative'
    plan = get_for_context(context, render_full_page)
    assert isinstance(plan, FewestFiles)

    # If we aren't rendering a full page, it should be SeparateEverything
    plan = get_for_context(context, False)
    assert isinstance(plan, SeparateEverything)

    settings.SKYLARK_PLANS = 'mediadeploy_bad'
    settings.SKYLARK_PLANS_DEFAULT = 'default'
    py.test.raises(ValueError, get_for_context, context, render_full_page)

    settings.DEBUG = False
    settings.SKYLARK_PLANS = 'mediadeploy_notthere'
    py.test.raises(MissingMediaPlan, get_for_context, context, render_full_page)


@with_setup(setup, teardown)
def test_deploy_reusable():
    hash_js1 = '96936cfbde05fe1cce209c7f344dbd08'
    hash_js2 = 'd921f357906a0e5e06da6a704376247b'
    hash_css = 'f27f963509901f608889f688b1e39f72'

    settings.DEBUG = False
    settings.SKYLARK_PLANS = 'mediadeploy_reusable'

    request = get_request_fixture()
    c = RequestContext(request)
    pa = PageAssembly('planapp/page/full.yaml', c)

    content = pa.dumps()

    jsfile = get_contents(
        os.path.join(cachedir, 'out', '%s.js' % hash_js1)
    )
    assert 'var static_uses2=null;var static_uses1=null;' in jsfile

    jsfile = get_contents(
        os.path.join(cachedir, 'out', '%s.js' % hash_js2)
    )
    assert jsfile.find('dojo.provide("dojox.timing._base")') < \
        jsfile.find('dojo.provide("dojox.timing")')
    assert jsfile.find("dojo.provide('PlanApp.Page.View')") < \
        jsfile.find("dojo.provide('PlanApp.Page.Controller')")
    assert jsfile.find("dojo.registerModulePath('ChirpTools'") < \
        jsfile.find("dojo.provide('ChirpTools.Error')")
    assert "dojo.registerModulePath('ChirpTools" not in content

    cssfile = get_contents(
        os.path.join(cachedir, 'out', '%s.css' % hash_css)
    )
    assert '.static_uses2' in cssfile
    assert '.static_uses1' in cssfile
    assert 'url(http://localhost:8000/media/' in cssfile
    assert 'out/planapp/page/media/img/uses1.gif' in cssfile

    assert '%s.js' % hash_js1 in content
    assert '%s.js' % hash_js2 in content
    assert 'planapp/page/media/js/static_uses1.js' not in content
    assert 'planapp/page/media/js/static_uses2.js' not in content

    assert '%s.css' % hash_css in content
    assert 'planapp/page/media/css/static_uses1.css' not in content
    assert 'planapp/page/media/css/static_uses2.css' not in content

    assert content.find('addon/dojo.js') < \
        content.find('%s.js' % hash_js1)
    assert content.find('%s.js' % hash_js1) < \
        content.find('media/uses1.js')

    assert content.find('media/uses1.css') < \
        content.find('%s.css' % hash_css)
    assert content.find('%s.css' % hash_css) < \
        content.find('site.com/handheld.css')

    assert content.find('js/ie7only.css') < \
        content.find('%s.js' % hash_js2)


@with_setup(setup, teardown)
def test_deploy_fewest():
    hash_css = 'a30e20a6a1d62976266b612a7e5d634a'
    hash_js = '99cd70ab43d662a64aa33c794433295a'

    settings.SKYLARK_PLANS = 'mediadeploy_fewest'

    request = get_request_fixture()
    c = RequestContext(request)
    pa = PageAssembly('planapp/page/full.yaml', c)

    content = pa.dumps()

    exist(
        'out/%s.css' % hash_css,
        'out/%s.js' % hash_js,
        'out/planapp/page/media/css/gte_ie6only.css',
        'out/planapp/page/media/img/uses1.gif',
        'out/planapp/page/media/img/uses2.gif',
        'out/planapp/page/media/js/Controller.js',
        'out/planapp/page/media/js/duplicated.js',
        'out/planapp/page/media/js/ie7only.js',
        'out/planapp/page/media/js/inline.js',
        'out/planapp/page/media/js/notreferenced.js',
        'out/planapp/page/media/js/static.js',
        'out/planapp/page/media/js/static_uses1.js',
        'out/planapp/page/media/js/static_uses2.js',
        'out/planapp/page/media/js/View.js',
    )

    jsfile = get_contents(
        os.path.join(cachedir, 'out', '%s.js' % hash_js)
    )
    assert 'static_uses2' in jsfile
    assert 'static_uses1' in jsfile
    assert 'static=' in jsfile
    assert 'duplicated' in jsfile
    assert jsfile.find('duplicated=null') < \
        jsfile.find('dojo.provide("dojox.timing._base")')

    cssfile = get_contents(
        os.path.join(cachedir, 'out', '%s.css' % hash_css)
    )

    assert '.static_uses2' in cssfile
    assert '.static_uses1' in cssfile
    assert '.static {' in cssfile

    assert '%s.js' % hash_js in content
    assert 'planapp/page/media/js/static' not in content
    assert 'planapp/page/media/js/duplicated' not in content

    assert '%s.css' % hash_css in content

    assert content.find('addon/dojo') < \
        content.find('%s.js' % hash_js)
    assert content.find('%s.js' % hash_js) < \
        content.find('media/uses1.js')

    assert content.find('media/uses1.css') < \
        content.find('%s.css' % hash_css)
    assert content.find('%s.css' % hash_css) < \
        content.find('site.com/handheld.css')


@with_setup(setup, teardown)
def test_deploy_fewest_instrumented():
    hash_js = 'cb0aa470231a819d45ee02783f5fc468'

    chirp.instrument_site(True)
    settings.SKYLARK_PLANS = 'mediadeploy_fewest'

    request = get_request_fixture()
    c = RequestContext(request)
    pa = PageAssembly('planapp/page/full.yaml', c)

    content = pa.dumps()

    assert '%s.js' % hash_js in content

    jsfile = get_contents(
        os.path.join(cachedir, 'out', '%s.js' % hash_js)
    )

    assert "dojo.provide('chirp')" in jsfile

    chirp.instrument_site(False)


@with_setup(setup, teardown)
def test_missing_rollup_requirement():
    settings.SKYLARK_PLANS = 'mediadeploy_fewest'

    request = get_request_fixture()
    c = RequestContext(request)
    pa = PageAssembly('planapp/page/full_missing.yaml', c)

    py.test.raises(TemplateDoesNotExist, pa.dumps)


@with_setup(setup, teardown)
def test_deploy_reusable_no_js_minifying():
    hash_js = 'd921f357906a0e5e06da6a704376247b'

    settings.SKYLARK_PLANS = 'mediadeploy_reusable'

    plan_options(minify_javascript=False)

    request = get_request_fixture()
    c = RequestContext(request)
    pa = PageAssembly('planapp/page/full.yaml', c)

    content = pa.dumps()

    jsfile = get_contents(
        os.path.join(cachedir, 'out', '%s.js' % hash_js)
    )

    assert '\n        ' in jsfile
    assert '//' in jsfile
    # Should be Dojo documentation in here
    assert '// summary:' in jsfile


@with_setup(setup, teardown)
def test_will_not_needlessly_rollup():
    settings.SKYLARK_PLANS = 'mediadeploy_reusable'

    hash_js1 = 'd921f357906a0e5e06da6a704376247b'
    filename = os.path.join(cachedir, 'out', '%s.js' % hash_js1)

    request = get_request_fixture()
    c = RequestContext(request)
    pa = PageAssembly('planapp/page/full.yaml', c)

    content = pa.dumps()

    first_time = os.stat(filename).st_mtime
    first_listing = os.listdir(os.path.join(cachedir, 'out'))

    sleep(1.0)

    request = get_request_fixture()
    c = RequestContext(request)
    pa = PageAssembly('planapp/page/full.yaml', c)

    content = pa.dumps()

    assert first_time == os.stat(filename).st_mtime
    assert first_listing == os.listdir(os.path.join(cachedir, 'out'))


@with_setup(setup, teardown)
def test_will_rollup_with_lessjs():
    settings.SKYLARK_PLANS = 'mediadeploy_reusable'

    request = get_request_fixture()
    c = RequestContext(request)
    pa = PageAssembly('planapp/page/lesscss.yaml', c)

    content = pa.dumps()

    assert 'cfcache/out/planapp/page/media/css/lessjs.css' in content
    assert '<link rel="stylesheet/less"' in content

    settings.SKYLARK_PLANS = 'mediadeploy_fewest'

    hash_css = '8015602c1f5583386d62d1478e2a8442'
    filename = os.path.join(cachedir, 'out', '%s.css' % hash_css)

    request = get_request_fixture()
    c = RequestContext(request)
    pa = PageAssembly('planapp/page/lesscss.yaml', c)

    content = pa.dumps()

    assert ('<link rel="stylesheet/less" type="text/css" '
        'href="http://localhost:8000/media/cfcache/out/'
        '%s.css" media="screen">' % hash_css in content)

    css_file = get_contents(filename)

    assert '@less' in css_file
    assert 'cfcache/out/planapp/page/media/img/header.png' in css_file


@with_setup(setup, teardown)
def test_will_rollup_with_lessc():
    settings.SKYLARK_PLANS = 'mediadeploy_reusable'

    request = get_request_fixture()
    c = RequestContext(request)
    pa = PageAssembly('planapp/page/lessc.yaml', c)

    content = pa.dumps()

    assert 'cfcache/out/planapp/page/media/css/lessjs.css' in content
    assert '<link rel="stylesheet"' in content

    settings.SKYLARK_PLANS = 'mediadeploy_fewest'

    hash_css = '8015602c1f5583386d62d1478e2a8442'
    filename = os.path.join(cachedir, 'out', '%s.css' % hash_css)

    request = get_request_fixture()
    c = RequestContext(request)
    pa = PageAssembly('planapp/page/lessc.yaml', c)

    content = pa.dumps()

    assert ('<link rel="stylesheet" type="text/css" '
        'href="http://localhost:8000/media/cfcache/out/'
        '%s.css" media="screen">' % hash_css in content)

    css_file = get_contents(filename)

    assert '@less' not in css_file
    assert 'cfcache/out/planapp/page/media/img/header.png' in css_file



MESSAGE_HELP = \
"""This command will create a Django Skylark page and kick start a configuration
file for you

Do something like this:
    ./manage.py skylarkpage -a goodiesapp -p list

Will create a page called "list" in the goodiesapp"""

MESSAGE_CONFIRM_TEMPLATE_CREATION = \
"""We've found your app but no templates directory.

We can create this for you and continue from here if you like.

Create %s (yes/no)? """

MESSAGE_CONFIRM_CREATION = """We're about to create these directories:

%s

And these files

%s

Are you sure you want to do this (yes/no)? """

CONTENT_YAML = \
"""title: Some title here for %(app)s
body: %(page_directory)s/%(page)s.html

#js:
    # Here are some examples (delete these if you don't use them)
    #- url: http://somewhere.com/js/javascript.js
    #- static: app/page/js/path.js

css:
    - static: %(page_directory)s/media/css/screen.css
      media: screen

    # Here are some examples (delete these if you don't use them)

    #- url: {{ MEDIA_URL }}/css/handheld.css
    #  media: all

    #- render: app/page/css/dynamic.css
    #  media: handheld, print, screen

    #- static: app/page/css/clevercss.css
    #  process: clevercss
    #  media: screen
"""

CONTENT_HTML = """<body>
    <h1>Your page is working (using only the finest baby frogs, dew picked and
    flown in from Iraq)</h1>
</body>"""

CONTENT_VIEW = """from django.http import HttpResponse
from skylark.page import PageAssembly, RequestContext

def %(page)s(request):
    c = RequestContext(request, {
        #'name': "value"
    })

    pa = PageAssembly('%(page_directory)s/%(page)s.yaml', c)

    return pa.get_http_response()
"""



from skylark.renderer import Renderer


class Html5(Renderer):
    doctype = '<!DOCTYPE html>'
    template_name = 'skylark/html5.html'
    snippet_template_name = 'skylark/htmlsnippet.html'


class Html401Strict(Renderer):
    doctype = '<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Strict//EN" "http://www.w3.org/TR/html4/strict.dtd">'
    template_name = 'skylark/html401.html'
    snippet_template_name = 'skylark/htmlsnippet.html'


class Html401Transitional(Renderer):
    doctype = '<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN" "http://www.w3.org/TR/html4/loose.dtd">'
    template_name = 'skylark/html401.html'
    snippet_template_name = 'skylark/htmlsnippet.html'


class Html401Frameset(Renderer):
    doctype = '<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Frameset//EN" "http://www.w3.org/TR/html4/frameset.dtd">'
    template_name = 'skylark/html401.html'
    snippet_template_name = 'skylark/htmlsnippet.html'



from base import BasePlan


class SeparateEverything(BasePlan):
    def prepare_js(self, page_instructions):
        if not hasattr(page_instructions, 'js'):
            return

        self._prepare_assets(page_instructions, ('media/js/templates',))

        self._prepare_file('js', page_instructions)

    def prepare_css(self, page_instructions):
        if not hasattr(page_instructions, 'css'):
            return

        self._prepare_assets(page_instructions, ('media/img',))

        self._prepare_file('css', page_instructions)



import py.test

from os.path import join
from nose.tools import with_setup
from nose.plugins.attrib import attr
from nose.plugins.skip import SkipTest
from django.test.client import Client
from django.core.urlresolvers import get_mod_func
from django.utils.importlib import import_module
from django.http import HttpResponseNotFound, HttpResponseServerError

from skylark.tests.urls import handler404, handler500
from skylark.tests import *

@with_setup(setup, teardown)
def test_handler_404():
    settings.DEBUG = False

    client = Client()
    response = client.get('/something/we/know/is/not/there')
    content = response.content

    assert isinstance(response, HttpResponseNotFound)
    assert '<h1>404</h1>' in content

@with_setup(setup, teardown)
def test_handler_500():
    settings.DEBUG = False
    mod_name, func_name = get_mod_func(handler500)
    callback = getattr(import_module(mod_name), func_name), {}
    response = callback[0](get_request_fixture())

    content = response.content

    assert isinstance(response, HttpResponseServerError)
    assert '<h1>500</h1>' in content



from time import time

from django.utils.importlib import import_module
from django.conf import settings


class ChirpError(Exception):
    pass

# Used to prevent getting too happy with the autodiscover
_loading = False

# Whether things are instrumented, in other words are we testing?
_instrumented = False

# When we started instrumenting the site, handy for shutting it down after
# a period of time expires
_time_instrumented = None

# List of Chirp modules that need to loaded
# This is to make sure that we don't try and import our Chirp testing modules
# before everything is in place (namely all the urlpatterns)
_chirp_modules = []


def check_instrumentation(meth):
    """
    Decorator to check the main settings and turn on instrumentation if needed
    """
    def check(*args, **kwargs):
        if settings.DEBUG and \
           settings.SKYLARK_CHIRP_INSTRUMENTED and \
           not is_instrumented():
            # The user has indicated that we need to be instrumented for
            # testing
            instrument_site(True)

        return meth(*args, **kwargs)
    return check


def is_instrumented():
    """
    Is the site currently instrumented for testing?
    """
    global _instrumented
    return _instrumented


def page_assembly_testcase_handler(page_instructions, renderer, assembly):
    """
    Used to add the testcase yaml file to the page instructions if we are being
    instrumented (testing)
    """
    assembly.add_page_instructions(
        page_instructions, 'chirp/testrunner/testcase.yaml')


def instrument_site(on):
    """
    Sets a bit within the application that tells every other view that we are
    testing the site using the Chirp test runner.
    """
    from skylark.assembly import BaseAssembly

    global _instrumented, _time_instrumented, _chirp_modules

    if _instrumented and on:
        raise ChirpError('Chirp is already instrumented and ready for testing, '
            'use is_instrumented to detect state')

    _instrumented = on
    _time_instrumented = time()

    if not on:
        # Take our testcase handler out of the loop
        BaseAssembly.unregister_handler(page_assembly_testcase_handler)
        # There is nothing else for us to do here
        return

    BaseAssembly.register_handler(page_assembly_testcase_handler)

    # Now we need to run through all the Chirp modules we found through
    # autodiscover and import them.  We didn't need any of it until we
    # instrumented our site
    for mod in _chirp_modules:
        import_module(mod)


def autodiscover():
    """
    Stolen directory from Django Admin.

    Auto-discover INSTALLED_APPS chirp.py modules and fail silently when
    not present.
    """
    global _loading
    if _loading:
        return
    _loading = True

    import imp
    from django.conf import settings

    global _chirp_modules

    for app in settings.INSTALLED_APPS:
        if app == 'skylark':
            # Skip Django Skylark, these aren't the droids we are looking for
            continue

        try:
            app_path = import_module(app).__path__
        except AttributeError:
            continue

        try:
            imp.find_module('chirp', app_path)
        except ImportError:
            continue

        _chirp_modules.append("%s.chirp" % app)
    _loading = False


class TestEntryPoint(object):
    def __init__(self, **kwargs):
        self.url = kwargs.get('url', None)
        self.name = kwargs.get('name', None)


class TestRegistry(object):
    _tests = []

    def add(self, url, **kwargs):
        # We don't want to add this if we aren't in DEBUG
        if not settings.DEBUG:
            return

        # We don't want duplicates, that won't help anyone
        if not isinstance(url, (str, unicode)):
            raise ChirpError('You must add URLs to the test registry, not %s' %
                type(url))

        if not url in self:
            self._tests.append(TestEntryPoint(
                url=url, name=kwargs.get('name', None)))

    def clear(self):
        for test in self._tests:
            self._tests.remove(test)

    def __contains__(self, key):
        url = key in [tep.url for tep in self._tests]
        name = key in [tep.name for tep in self._tests]
        return url | name

    def __len__(self):
        return len(self._tests)

    def list(self):
        return self._tests


test_registry = TestRegistry()



from django.conf import settings


def media(request):
    from skylark.plans import get_plan
    plan = get_plan(settings.SKYLARK_PLANS,
                    settings.SKYLARK_PLANS_DEFAULT)
    return {
        'MEDIA_URL': settings.MEDIA_URL,
        'CF_MEDIA_URL': '%s%s/' % (
            settings.SKYLARK_CACHE_URL, plan.cache_prefix,)}



import time

from django.conf import settings
from django.template import loader, TemplateDoesNotExist
from django.http import HttpResponse
from django.core.urlresolvers import reverse
from django.utils import simplejson 

from skylark import RequestContext
from skylark.page import PageAssembly
from skylark.chirp import test_registry, is_instrumented, instrument_site


def testrunner(request):
    """
    Main view to load the interface and start running the tests
    """
    return interface_start(request)


def interface_start(request):
    """
    Provides the main HTML for the test runner
    """
    if not is_instrumented():
        instrument_site(True)
    teps = [ i.__dict__ for i in test_registry.list() ]
    context = RequestContext(request, {
        'test_count': len(test_registry),
        'test_entry_points_json': simplejson.dumps(teps),
        'url_deinstrument': reverse(deinstrument),
    })

    pa = PageAssembly('chirp/testrunner/display/display.yaml', context)
    return pa.get_http_response()


def deinstrument(request):
    """
    Called at the end of the test run, de-instuments the site for testing
    """
    instrument_site(False)
    return HttpResponse()


def subject_start(request):
    """
    Is the temporary page we see as the test runner is bootstrapping, it gives
    us some brief information before the test runs start
    """
    context = RequestContext(request)
    pa = PageAssembly('chirp/testrunner/subjectstart/subjectstart.yaml', context)
    return pa.get_http_response()

# Don't test this module
setattr(testrunner, '__test__', False)



from skylark.processor import clevercss

def test_webkit_properties():
    property ="""div:
    background-image: -webkit-gradient(linear, 0% 0%, 0% 100%, from(#E0E0E0), to(#C4C4C4))"""

    expected = """div {
  background-image: -webkit-gradient(linear, 0% 0%, 0% 100%, from(#E0E0E0), to(#C4C4C4));
}"""

    assert clevercss.convert(property) == expected 



# -*- coding: utf-8 -*-
"""
    CleverCSS
    ~~~~~~~~~

    The Pythonic way of CSS files.

    To convert this into a normal css file just call the `convert`
    function in the clevercss module. It's that easy :-)

    Example::

        base_padding = 2px
        background_color = #eee
        text_color = #111
        link_color = #ff0000

        body:
            font-family: serif, sans-serif, 'Verdana', 'Times New Roman'
            color: $text_color
            padding->
                top: $base_padding + 2
                right: $base_padding + 3
                left: $base_padding + 3
                bottom: $base_padding + 2
            background-color: $background_color

        div.foo:
            width: "Hello World".length() * 20px
            foo: (foo, bar, baz, 42).join('/')

        a:
            color: $link_color
            &:hover:
                color: $link_color.darken(30%)
            &:active:
                color: $link_color.brighten(10%)

        div.navigation:
            height: 1.2em
            padding: 0.2em
            ul:
                margin: 0
                padding: 0
                list-style: none
                li:
                    float: left
                    height: 1.2em
                    a:
                        display: block
                        height: 1em
                        padding: 0.1em
            foo: (1 2 3).string()

        __END__
        this is ignored, but __END__ as such is completely optional.

    To get the converted example module as css just run this file as script
    with the "--eigen-test" parameter.

    Literals
    --------

    CleverCSS supports most of the standard CSS literals.  Some syntax
    elements are not supported by now, some will probably never.

    Strings:
        everything (except of dangling method calls and whitespace) that
        cannot be parsed with a different rule is considered being a
        string.  If you want to have whitespace in your strings or use
        something as string that would otherwise have a different semantic
        you can use double or single quotes.

        these are all valid strings::

            =
            foo-bar-baz
            "blub"
            'foo bar baz'
            Verdana

    Numbers
        Numbers are just that.  Numbers with unit postfix are values.

    Values
        Values are numbers with an associated unit.  Most obvious difference
        between those two are the different semantics in arithmetic
        operations.  Some units can be converted, some are just not compatible
        (for example you won't be able to convert 1em in percent because
         there is no fixed conversion possible)
        Additionally to the CSS supported colors this module supports the
        netscape color codes.

    Colors
        Colors are so far only supported in hexadecimal notation.  You can
        also use the `rgb()` literal to some amount.  But that means you
        cannot use "orange" as color.

    URLs:
        URLs work like strings, the only difference is that the syntax looks
        like ``url(...)``.

    Variables:
        variables are quite simple.  Once they are defined in the root section
        you can use them in every expression::

            foo = 42px

            div:
                width: $foo * 100;

    Lists:
        Sometimes you want to assign more than one element to a CSS rule.  For
        example if you work with font families.  In that situation just use
        the comma operator to define a list::

            font-family: Verdana, Arial, sans-serif

        Additionally lists have methods, you can for example do this (although
        probably completely useless in real world cases)::

            width: (1, 2, 3, 4).length() * 20


    Implicit Concatenation
    ----------------------

    CleverCSS ignores whitespace.  But whitespace keeps the tokens apart.  If
    the parser now stumbles upon something it doesn't know how to handle, it
    assumes that there was a whitespace.  In some situations CSS even requires
    that behavior::

        padding: 2px 3px

    But because CleverCSS has expressions this could lead to this situation::

        padding: $x + 1 $x + 2

    This if course works too because ``$x + 1`` is one expression and
    ``$x + 2`` another one.  This however can lead to code that is harder to
    read.  In that situation it's recommended to parentize the expressions::

        padding: ($x + 1) ($x + 2)

    or remove the whitespace between the operators::

        padding: $x+1 $x+2


    Operators
    ---------

    ``+``       add two numbers, a number and a value or two compatible
                values (for example ``1cm + 12mm``).  This also works as
                concatenate operator for strings.  Using this operator
                on color objects allows some basic color composition.
    ``-``       subtract one number from another, a number from a value
                or a value from a compatible one.  Like the plus operator
                this also works on colors.
    ``*``       Multiply numbers, numbers with a value.  Multiplying strings
                repeats it. (eg: ``= * 5`` gives '=====')
    ``/``       divide one number or value by a number.
    ``%``       do a modulo division on a number or value by a number.

    Keep in mind that whitespace matters. For example ``20% 10`` is something
    completely different than ``20 % 10``. The first one is an implicit
    concatenation expression with the values 20% and 10, the second one a
    modulo epression.  The same applies to ``no-wrap`` versus ``no - wrap``
    and others.

    Additionally there are two operators used to keep list items apart. The
    comma (``,``) and semicolon (``;``) operator both keep list items apart.

    If you want to group expressions you can use parentheses.

    Methods
    -------

    Objects have some methods you can call:

    - `Number.abs()`            get the absolute value of the number
    - `Number.round(places)`    round to (default = 0) places
    - `Value.abs()`             get the absolute value for this value
    - `Value.round(places)`     round the value to (default = 0) places
    - `Color.brighten(amount)`  brighten the color by amount percent of
                                the current lightness, or by 0 - 100.
                                brightening by 100 will result in white.
    - `Color.darken(amount)`    darken the color by amount percent of the
                                current lightness, or by 0 - 100.
                                darkening by 100 will result in black.
    - `String.length()`         the length of the string.
    - `String.upper()`          uppercase version of the string.
    - `String.lower()`          lowercase version of the string.
    - `String.strip()`          version with leading an trailing whitespace
                                removed.
    - `String.split(delim)`     return a list of substrings, splitted by
                                whitespace or delim.
    - `String.eval()`           eval a css rule inside of a string. For
                                example a string "42" would return the
                                number 42 when parsed. But this can also
                                contain complex expressions such as
                                "(1 + 2) * 3px".
    - `String.string()`         just return the string itself.
    - `List.length()`           number of elements in a list.
    - `List.join(delim)`        join a list by space char or delim.

    Additionally all objects and expressions have a `.string()` method that
    converts the object into a string, and a `.type()` method that returns
    the type of the object as string.

    If you have implicit concatenated expressions you can convert them into
    a list using the `list` method::

        (1 2 3 4 5).list()

    does the same as::

        1, 2, 3, 4, 5

    :copyright: Copyright 2007 by Armin Ronacher, Georg Brandl.
    :license: BSD License
"""
import re
import colorsys
import operator


VERSION = '0.1'

__all__ = ['convert']


# regular expresssions for the normal parser
_var_def_re = re.compile(r'^([a-zA-Z_][a-zA-Z0-9_]*)\s*=\s*(.+)')
_def_re = re.compile(r'^([a-zA-Z-]+(?:\s*,\s*[a-zA-Z-]+)*)\s*:\s*(.+)')
_line_comment_re = re.compile(r'//.*?$')

# list of operators
_operators = ['+', '-', '*', '/', '%', '(', ')', ';', ',']

# units and conversions
_units = ['em', 'ex', 'px', 'cm', 'mm', 'in', 'pt', 'pc', 'deg', 'rad'
          'grad', 'ms', 's', 'Hz', 'kHz', '%']
_conv = {
    'length': {
        'mm':       1.0,
        'cm':       10.0,
        'in':       25.4,
        'pt':       25.4 / 72,
        'pc':       25.4 / 6
    },
    'time': {
        'ms':       1.0,
        's':        1000.0
    },
    'freq': {
        'Hz':       1.0,
        'kHz':      1000.0
    }
}
_conv_mapping = {}
for t, m in _conv.iteritems():
    for k in m:
        _conv_mapping[k] = t
del t, m, k

# color literals
_colors = {
    'aliceblue': '#f0f8ff',
    'antiquewhite': '#faebd7',
    'aqua': '#00ffff',
    'aquamarine': '#7fffd4',
    'azure': '#f0ffff',
    'beige': '#f5f5dc',
    'bisque': '#ffe4c4',
    'black': '#000000',
    'blanchedalmond': '#ffebcd',
    'blue': '#0000ff',
    'blueviolet': '#8a2be2',
    'brown': '#a52a2a',
    'burlywood': '#deb887',
    'cadetblue': '#5f9ea0',
    'chartreuse': '#7fff00',
    'chocolate': '#d2691e',
    'coral': '#ff7f50',
    'cornflowerblue': '#6495ed',
    'cornsilk': '#fff8dc',
    'crimson': '#dc143c',
    'cyan': '#00ffff',
    'darkblue': '#00008b',
    'darkcyan': '#008b8b',
    'darkgoldenrod': '#b8860b',
    'darkgray': '#a9a9a9',
    'darkgreen': '#006400',
    'darkkhaki': '#bdb76b',
    'darkmagenta': '#8b008b',
    'darkolivegreen': '#556b2f',
    'darkorange': '#ff8c00',
    'darkorchid': '#9932cc',
    'darkred': '#8b0000',
    'darksalmon': '#e9967a',
    'darkseagreen': '#8fbc8f',
    'darkslateblue': '#483d8b',
    'darkslategray': '#2f4f4f',
    'darkturquoise': '#00ced1',
    'darkviolet': '#9400d3',
    'deeppink': '#ff1493',
    'deepskyblue': '#00bfff',
    'dimgray': '#696969',
    'dodgerblue': '#1e90ff',
    'firebrick': '#b22222',
    'floralwhite': '#fffaf0',
    'forestgreen': '#228b22',
    'fuchsia': '#ff00ff',
    'gainsboro': '#dcdcdc',
    'ghostwhite': '#f8f8ff',
    'gold': '#ffd700',
    'goldenrod': '#daa520',
    'gray': '#808080',
    'green': '#008000',
    'greenyellow': '#adff2f',
    'honeydew': '#f0fff0',
    'hotpink': '#ff69b4',
    'indianred': '#cd5c5c',
    'indigo': '#4b0082',
    'ivory': '#fffff0',
    'khaki': '#f0e68c',
    'lavender': '#e6e6fa',
    'lavenderblush': '#fff0f5',
    'lawngreen': '#7cfc00',
    'lemonchiffon': '#fffacd',
    'lightblue': '#add8e6',
    'lightcoral': '#f08080',
    'lightcyan': '#e0ffff',
    'lightgoldenrodyellow': '#fafad2',
    'lightgreen': '#90ee90',
    'lightgrey': '#d3d3d3',
    'lightpink': '#ffb6c1',
    'lightsalmon': '#ffa07a',
    'lightseagreen': '#20b2aa',
    'lightskyblue': '#87cefa',
    'lightslategray': '#778899',
    'lightsteelblue': '#b0c4de',
    'lightyellow': '#ffffe0',
    'lime': '#00ff00',
    'limegreen': '#32cd32',
    'linen': '#faf0e6',
    'magenta': '#ff00ff',
    'maroon': '#800000',
    'mediumaquamarine': '#66cdaa',
    'mediumblue': '#0000cd',
    'mediumorchid': '#ba55d3',
    'mediumpurple': '#9370db',
    'mediumseagreen': '#3cb371',
    'mediumslateblue': '#7b68ee',
    'mediumspringgreen': '#00fa9a',
    'mediumturquoise': '#48d1cc',
    'mediumvioletred': '#c71585',
    'midnightblue': '#191970',
    'mintcream': '#f5fffa',
    'mistyrose': '#ffe4e1',
    'moccasin': '#ffe4b5',
    'navajowhite': '#ffdead',
    'navy': '#000080',
    'oldlace': '#fdf5e6',
    'olive': '#808000',
    'olivedrab': '#6b8e23',
    'orange': '#ffa500',
    'orangered': '#ff4500',
    'orchid': '#da70d6',
    'palegoldenrod': '#eee8aa',
    'palegreen': '#98fb98',
    'paleturquoise': '#afeeee',
    'palevioletred': '#db7093',
    'papayawhip': '#ffefd5',
    'peachpuff': '#ffdab9',
    'peru': '#cd853f',
    'pink': '#ffc0cb',
    'plum': '#dda0dd',
    'powderblue': '#b0e0e6',
    'purple': '#800080',
    'red': '#ff0000',
    'rosybrown': '#bc8f8f',
    'royalblue': '#4169e1',
    'saddlebrown': '#8b4513',
    'salmon': '#fa8072',
    'sandybrown': '#f4a460',
    'seagreen': '#2e8b57',
    'seashell': '#fff5ee',
    'sienna': '#a0522d',
    'silver': '#c0c0c0',
    'skyblue': '#87ceeb',
    'slateblue': '#6a5acd',
    'slategray': '#708090',
    'snow': '#fffafa',
    'springgreen': '#00ff7f',
    'steelblue': '#4682b4',
    'tan': '#d2b48c',
    'teal': '#008080',
    'thistle': '#d8bfd8',
    'tomato': '#ff6347',
    'turquoise': '#40e0d0',
    'violet': '#ee82ee',
    'wheat': '#f5deb3',
    'white': '#ffffff',
    'whitesmoke': '#f5f5f5',
    'yellow': '#ffff00',
    'yellowgreen': '#9acd32'
}
_reverse_colors = dict((v, k) for k, v in _colors.iteritems())

# partial regular expressions for the expr parser
_r_number = '\d+(?:\.\d+)?'
_r_string = r"(?:'(?:[^'\\]*(?:\\.[^'\\]*)*)'|" \
            r'\"(?:[^"\\]*(?:\\.[^"\\]*)*)")'
_r_call = r'([a-zA-Z_][a-zA-Z0-9_]*)\('

# regular expressions for the expr parser
_operator_re = re.compile('|'.join(re.escape(x) for x in _operators))
_whitespace_re = re.compile(r'\s+')
_number_re = re.compile(_r_number + '(?![a-zA-Z0-9_])')
_number_re = re.compile(_r_number + '(?![a-zA-Z0-9_])')
_value_re = re.compile(r'(%s)(%s)(?![a-zA-Z0-9_])' % (_r_number, '|'.join(_units)))
_color_re = re.compile(r'#' + ('[a-fA-f0-9]{1,2}' * 3))
_string_re = re.compile('%s|([^\s*/();,.+$]+|\.(?!%s))+' % (_r_string, _r_call))
_webkit_re = re.compile(r'\-webkit\-(.*)$')
_url_re = re.compile(r'url\(\s*(%s|.*?)\s*\)' % _r_string)
_rgba_re = re.compile(r'rgba\(\s*(%s|.*?)\s*\)' % _r_string)
_var_re = re.compile(r'(?<!\\)\$(?:([a-zA-Z_][a-zA-Z0-9_]*)|'
                     r'\{([a-zA-Z_][a-zA-Z0-9_]*)\})')
_call_re = re.compile(r'\.' + _r_call)


def number_repr(value):
    """
    CleverCSS uses floats internally.  To keep the string representation
    of the numbers small cut off the places if this is possible without
    loosing much information.
    """
    value = unicode(value)
    parts = value.rsplit('.')
    if len(parts) == 2 and parts[-1] == '0':
        return parts[0]
    return value


def rgb_to_hls(red, green, blue):
    """
    Convert RGB to HSL.  The RGB values we use are in the range 0-255, but
    HSL is in the range 0-1!
    """
    return colorsys.rgb_to_hls(red / 255.0, green / 255.0, blue / 255.0)


def hls_to_rgb(hue, saturation, lightness):
    """Convert HSL back to RGB."""
    t = colorsys.hls_to_rgb(hue, saturation, lightness)
    return tuple(int(x * 255) for x in t)


class ParserError(Exception):
    """
    Raised on syntax errors.
    """

    def __init__(self, lineno, message):
        self.lineno = lineno
        Exception.__init__(self, message)

    def __str__(self):
        return '%s (line %s)' % (
            self.message,
            self.lineno
        )


class EvalException(Exception):
    """
    Raised during evaluation.
    """

    def __init__(self, lineno, message):
        self.lineno = lineno
        Exception.__init__(self, message)

    def __str__(self):
        return '%s (line %s)' % (
            self.message,
            self.lineno
        )


class LineIterator(object):
    """
    This class acts as an iterator for sourcecode. It yields the lines
    without comments or empty lines and keeps track of the real line
    number.

    Example::

        >>> li = LineIterator(u'foo\nbar\n\n/* foo */bar')
        >>> li.next()
        1, u'foo'
        >>> li.next()
        2, 'bar'
        >>> li.next()
        4, 'bar'
        >>> li.next()
        Traceback (most recent call last):
          File "<stdin>", line 1, in <module>
        StopIteration
    """

    def __init__(self, source, emit_endmarker=False):
        """
        If `emit_endmarkers` is set to `True` the line iterator will send
        the string ``'__END__'`` before closing down.
        """
        lines = source.splitlines()
        self.lineno = 0
        self.lines = len(lines)
        self.emit_endmarker = emit_endmarker
        self._lineiter = iter(lines)

    def __iter__(self):
        return self

    def _read_line(self):
        """Read the next non empty line.  This strips line comments."""
        line = ''
        while not line.strip():
            line += _line_comment_re.sub('', self._lineiter.next()).rstrip()
            self.lineno += 1
        return line

    def _next(self):
        """
        Get the next line without mutliline comments.
        """
        # XXX: this fails for a line like this: "/* foo */bar/*"
        line = self._read_line()
        comment_start = line.find('/*')
        if comment_start < 0:
            return self.lineno, line

        stripped_line = line[:comment_start]
        comment_end = line.find('*/', comment_start)
        if comment_end >= 0:
            return self.lineno, stripped_line + line[comment_end + 2:]

        start_lineno = self.lineno
        try:
            while True:
                line = self._read_line()
                comment_end = line.find('*/')
                if comment_end >= 0:
                    stripped_line += line[comment_end + 2:]
                    break
        except StopIteration:
            raise ParserError(self.lineno, 'missing end of multiline comment')
        return start_lineno, stripped_line

    def next(self):
        """
        Get the next line without multiline comments and emit the
        endmarker if we reached the end of the sourcecode and endmarkers
        were requested.
        """
        try:
            return self._next()
        except StopIteration:
            if self.emit_endmarker:
                self.emit_endmarker = False
                return self.lineno, '__END__'
            raise


class Engine(object):
    """
    The central object that brings parser and evaluation together.  Usually
    nobody uses this because the `convert` function wraps it.
    """

    def __init__(self, source):
        self._parser = p = Parser()
        self.rules, self._vars = p.parse(source)

    def evaluate(self, context=None):
        """Evaluate code."""
        expr = None
        context = {}
        for key, value in context.iteritems():
            expr = self._parser.parse_expr(1, value)
            context[key] = expr
        context.update(self._vars)

        for selectors, defs in self.rules:
            yield selectors, [(key, expr.to_string(context))
                              for key, expr in defs]

    def to_css(self, context=None):
        """Evaluate the code and generate a CSS file."""
        blocks = []
        for selectors, defs in self.evaluate(context):
            block = []
            block.append(u',\n'.join(selectors) + ' {')
            for key, value in defs:
                block.append(u'  %s: %s;' % (key, value))
            block.append('}')
            blocks.append(u'\n'.join(block))
        return u'\n\n'.join(blocks)


class TokenStream(object):
    """
    This is used by the expression parser to manage the tokens.
    """

    def __init__(self, lineno, gen):
        self.lineno = lineno
        self.gen = gen
        self.next()

    def next(self):
        try:
            self.current = self.gen.next()
        except StopIteration:
            self.current = None, 'eof'

    def expect(self, value, token):
        if self.current != (value, token):
            raise ParserError(self.lineno, "expected '%s', got '%s'." %
                              (value, self.current[0]))
        self.next()


class Expr(object):
    """
    Baseclass for all expressions.
    """

    #: name for exceptions
    name = 'expression'

    #: empty iterable of dict with methods
    methods = ()

    def __init__(self, lineno=None):
        self.lineno = lineno

    def evaluate(self, context):
        return self

    def add(self, other, context):
        return String(self.to_string(context) + other.to_string(context))

    def sub(self, other, context):
        raise EvalException(self.lineno, 'cannot substract %s from %s' %
                            (self.name, other.name))

    def mul(self, other, context):
        raise EvalException(self.lineno, 'cannot multiply %s with %s' %
                            (self.name, other.name))

    def div(self, other, context):
        raise EvalException(self.lineno, 'cannot divide %s by %s' %
                            (self.name, other.name))

    def mod(self, other, context):
        raise EvalException(self.lineno, 'cannot use the modulo operator for '
                            '%s and %s. Misplaced unit symbol?' %
                            (self.name, other.name))

    def neg(self, context):
        raise EvalException(self.lineno, 'cannot negate %s by %s' % self.name)

    def to_string(self, context):
        return self.evaluate(context).to_string(context)

    def call(self, name, args, context):
        if name == 'string':
            if isinstance(self, String):
                return self
            return String(self.to_string(context))
        elif name == 'type':
            return String(self.name)
        if name not in self.methods:
            raise EvalException(self.lineno, '%s objects don\'t have a method'
                                ' called "%s". If you want to use this'
                                ' construct as string, quote it.' %
                                (self.name, name))
        return self.methods[name](self, context, *args)

    def __repr__(self):
        return '%s(%s)' % (
            self.__class__.__name__,
            ', '.join('%s=%r' % item for item in
                      self.__dict__.iteritems())
        )


class ImplicitConcat(Expr):
    """
    Holds multiple expressions that are delimited by whitespace.
    """
    name = 'concatenated'
    methods = {
        'list':     lambda x, c: List(x.nodes)
    }

    def __init__(self, nodes, lineno=None):
        Expr.__init__(self, lineno)
        self.nodes = nodes

    def to_string(self, context):
        return u' '.join(x.to_string(context) for x in self.nodes)


class Bin(Expr):

    def __init__(self, left, right, lineno=None):
        Expr.__init__(self, lineno)
        self.left = left
        self.right = right


class Add(Bin):

    def evaluate(self, context):
        return self.left.evaluate(context).add(
               self.right.evaluate(context), context)


class Sub(Bin):

    def evaluate(self, context):
        return self.left.evaluate(context).sub(
               self.right.evaluate(context), context)


class Mul(Bin):

    def evaluate(self, context):
        return self.left.evaluate(context).mul(
               self.right.evaluate(context), context)


class Div(Bin):

    def evaluate(self, context):
        return self.left.evaluate(context).div(
               self.right.evaluate(context), context)


class Mod(Bin):

    def evaluate(self, context):
        return self.left.evaluate(context).mod(
               self.right.evaluate(context), context)


class Neg(Expr):

    def __init__(self, node, lineno=None):
        Expr.__init__(self, lineno)
        self.node = node

    def evaluate(self, context):
        return self.node.evaluate(context).neg(context)


class Call(Expr):

    def __init__(self, node, method, args, lineno=None):
        Expr.__init__(self, lineno)
        self.node = node
        self.method = method
        self.args = args

    def evaluate(self, context):
        return self.node.evaluate(context) \
                        .call(self.method, [x.evaluate(context)
                                            for x in self.args],
                              context)


class Literal(Expr):

    def __init__(self, value, lineno=None):
        Expr.__init__(self, lineno)
        self.value = value

    def to_string(self, context):
        rv = unicode(self.value)
        if len(rv.split(None, 1)) > 1:
            return u"'%s'" % rv.replace('\\', '\\\\') \
                               .replace('\n', '\\\n') \
                               .replace('\t', '\\\t') \
                               .replace('\'', '\\\'')
        return rv


class Number(Literal):
    name = 'number'

    methods = {
        'abs':      lambda x, c: Number(abs(x.value)),
        'round':    lambda x, c, p=0: Number(round(x.value, p))
    }

    def __init__(self, value, lineno=None):
        Literal.__init__(self, float(value), lineno)

    def add(self, other, context):
        if isinstance(other, Number):
            return Number(self.value + other.value, lineno=self.lineno)
        elif isinstance(other, Value):
            return Value(self.value + other.value, other.unit,
                         lineno=self.lineno)
        return Literal.add(self, other, context)

    def sub(self, other, context):
        if isinstance(other, Number):
            return Number(self.value - other.value, lineno=self.lineno)
        elif isinstance(other, Value):
            return Value(self.value - other.value, other.unit,
                         lineno=self.lineno)
        return Literal.sub(self, other, context)

    def mul(self, other, context):
        if isinstance(other, Number):
            return Number(self.value * other.value, lineno=self.lineno)
        elif isinstance(other, Value):
            return Value(self.value * other.value, other.unit,
                         lineno=self.lineno)
        return Literal.mul(self, other, context)

    def div(self, other, context):
        try:
            if isinstance(other, Number):
                return Number(self.value / other.value, lineno=self.lineno)
            elif isinstance(other, Value):
                return Value(self.value / other.value, other.unit,
                             lineno=self.lineno)
            return Literal.div(self, other, context)
        except ZeroDivisionError:
            raise EvalException(self.lineno, 'cannot divide by zero')

    def mod(self, other, context):
        try:
            if isinstance(other, Number):
                return Number(self.value % other.value, lineno=self.lineno)
            elif isinstance(other, Value):
                return Value(self.value % other.value, other.unit,
                             lineno=self.lineno)
            return Literal.mod(self, other, context)
        except ZeroDivisionError:
            raise EvalException(self.lineno, 'cannot divide by zero')

    def neg(self, context):
        return Number(-self.value)

    def to_string(self, context):
        return number_repr(self.value)


class Value(Literal):
    name = 'value'

    methods = {
        'abs':      lambda x, c: Value(abs(x.value), x.unit),
        'round':    lambda x, c, p=0: Value(round(x.value, p), x.unit)
    }

    def __init__(self, value, unit, lineno=None):
        Literal.__init__(self, float(value), lineno)
        self.unit = unit

    def add(self, other, context):
        return self._conv_calc(other, context, operator.add, Literal.add,
                               'cannot add %s and %s')

    def sub(self, other, context):
        return self._conv_calc(other, context, operator.sub, Literal.sub,
                               'cannot subtract %s from %s')

    def mul(self, other, context):
        if isinstance(other, Number):
            return Value(self.value * other.value, self.unit,
                         lineno=self.lineno)
        return Literal.mul(self, other, context)

    def div(self, other, context):
        if isinstance(other, Number):
            try:
                return Value(self.value / other.value, self.unit,
                             lineno=self.lineno)
            except ZeroDivisionError:
                raise EvalException(self.lineno, 'cannot divide by zero',
                                    lineno=self.lineno)
        return Literal.div(self, other, context)

    def mod(self, other, context):
        if isinstance(other, Number):
            try:
                return Value(self.value % other.value, self.unit,
                             lineno=self.lineno)
            except ZeroDivisionError:
                raise EvalException(self.lineno, 'cannot divide by zero')
        return Literal.mod(self, other, context)

    def _conv_calc(self, other, context, calc, fallback, msg):
        if isinstance(other, Number):
            return Value(calc(self.value, other.value), self.unit)
        elif isinstance(other, Value):
            if self.unit == other.unit:
                return Value(self.value + other.value, other.unit,
                             lineno=self.lineno)
            self_unit_type = _conv_mapping.get(self.unit)
            other_unit_type = _conv_mapping.get(other.unit)
            if not self_unit_type or not other_unit_type or \
               self_unit_type != other_unit_type:
                raise EvalException(self.lineno, msg % (self.unit, other.unit)
                                    + ' because the two units are '
                                    'not compatible.')
            self_unit = _conv[self_unit_type][self.unit]
            other_unit = _conv[other_unit_type][other.unit]
            if self_unit > other_unit:
                return Value(calc(self.value / other_unit * self_unit,
                                  other.value), other.unit,
                             lineno=self.lineno)
            return Value(calc(other.value / self_unit * other_unit,
                              self.value), self.unit, lineno=self.lineno)
        return fallback(self, other, context)

    def neg(self, context):
        return Value(-self.value, self.unit, lineno=self.lineno)

    def to_string(self, context):
        return number_repr(self.value) + self.unit

def adjust_color(color, amount=None, attribute=None):
    if amount is None:
        amount = Value(10.0, '%')
    hls = {}
    hls['hue'], hls['lightness'], hls['saturation'] = rgb_to_hls(*color.value)
    if isinstance(amount, Value):
        if amount.unit == '%':
            if not amount.value:
                return color
            hls[attribute] *= amount.value / 100.0
        else:
            raise EvalException(self.lineno, 'invalid unit %s for color '
                                'calculations.' % amount.unit)
    elif isinstance(amount, Number):
        hls[attribute] += (amount.value / 100.0)
    
    if attribute == 'hue':
        while hls[attribute] < 0 or hls[attribute] > 1.0:
            if hls[attribute] < 0:
                hls[attribute] += 1.0
            elif hls[attribute] > 0:
                hls[attribute] -= 1.0
    else:
        if hls[attribute] < 0:
            hls[attribute] = 0.0
        elif hls[attribute] > 1:
            hls[attribute] = 1.0

    return Color(hls_to_rgb(hls['hue'], hls['lightness'], hls['saturation']))

def brighten_color(color, context, amount=None):
    return adjust_color(color, amount, 'lightness')

def darken_color(color, context, amount=None):
    # Since we are darkening, this is a sign flip on the value
    if isinstance(amount, Number):
        amount.value *= -1

    return adjust_color(color, amount, 'lightness')

def saturate_color(color, context, amount=None):
    return adjust_color(color, amount, 'saturation')

def shift_color(color, context, amount=None):
    return adjust_color(color, amount, 'hue')

class Color(Literal):
    name = 'color'

    methods = {
        'brighten': brighten_color,
        'darken':   darken_color,
        'shift':    shift_color,
        'saturate': saturate_color,
        'hex':      lambda x, c: Color(x.value, x.lineno)
    }

    def __init__(self, value, lineno=None):
        self.from_name = False
        if isinstance(value, basestring):
            if not value.startswith('#'):
                value = _colors.get(value)
                if not value:
                    raise ParserError(lineno, 'unknown color name')
                self.from_name = True
            try:
                if len(value) == 4:
                    value = [int(x * 2, 16) for x in value[1:]]
                elif len(value) == 7:
                    value = [int(value[i:i + 2], 16) for i in xrange(1, 7, 2)]
                else:
                    raise ValueError()
            except ValueError, e:
                raise ParserError(lineno, 'invalid color value')
        Literal.__init__(self, tuple(value), lineno)

    def add(self, other, context):
        if isinstance(other, (Color, Number)):
            return self._calc(other, operator.add)
        return Literal.add(self, other, context)

    def sub(self, other, context):
        if isinstance(other, (Color, Number)):
            return self._calc(other, operator.sub)
        return Literal.sub(self, other, context)

    def mul(self, other, context):
        if isinstance(other, (Color, Number)):
            return self._calc(other, operator.mul)
        return Literal.mul(self, other, context)

    def div(self, other, context):
        if isinstance(other, (Color, Number)):
            return self._calc(other, operator.sub)
        return Literal.div(self, other, context)

    def to_string(self, context):
        code = '#%02x%02x%02x' % self.value
        return self.from_name and _reverse_colors.get(code) or code

    def _calc(self, other, method):
        is_number = isinstance(other, Number)
        channels = []
        for idx, val in enumerate(self.value):
            if is_number:
                other_val = int(other.value)
            else:
                other_val = other.value[idx]
            new_val = method(val, other_val)
            if new_val > 255:
                new_val = 255
            elif new_val < 0:
                new_val = 0
            channels.append(new_val)
        return Color(tuple(channels), lineno=self.lineno)


class RGB(Expr):
    """
    an expression that hopefully returns a Color object.
    """

    def __init__(self, rgb, lineno=None):
        Expr.__init__(self, lineno)
        self.rgb = rgb

    def evaluate(self, context):
        args = []
        for arg in self.rgb:
            arg = arg.evaluate(context)
            if isinstance(arg, Number):
                value = int(arg.value)
            elif isinstance(arg, Value) and arg.unit == '%':
                value = int(arg.value / 100.0 * 255)
            else:
                raise EvalException(self.lineno, 'colors defined using the '
                                    'rgb() literal only accept numbers and '
                                    'percentages.')
            if value < 0 or value > 255:
                raise EvalError(self.lineno, 'rgb components must be in '
                                'the range 0 to 255.')
            args.append(value)
        return Color(args, lineno=self.lineno)


class String(Literal):
    name = 'string'

    methods = {
        'length':   lambda x, c: Number(len(x.value)),
        'upper':    lambda x, c: String(x.value.upper()),
        'lower':    lambda x, c: String(x.value.lower()),
        'strip':    lambda x, c: String(x.value.strip()),
        'split':    lambda x, c, d=None: String(x.value.split(d)),
        'eval':     lambda x, c: Parser().parse_expr(x.lineno, x.value)
                                         .evaluate(c)
    }

    def mul(self, other, context):
        if isinstance(other, Number):
            return String(self.value * int(other.value), lineno=self.lineno)
        return Literal.mul(self, other, context, lineno=self.lineno)

class Webkit(Literal):
    name = 'Webkit'

    def to_string(self, context):
        return '-webkit-%s' % self.value

class RGBA(Literal):
    name = 'RGBA'

    def to_string(self, context):
        return 'rgba(%s)' % self.value

class URL(Literal):
    name = 'URL'
    methods = {
        'length':   lambda x, c: Number(len(self.value))
    }

    def add(self, other, context):
        return URL(self.value + other.to_string(context),
                   lineno=self.lineno)

    def mul(self, other, context):
        if isinstance(other, Number):
            return URL(self.value * int(other.value), lineno=self.lineno)
        return Literal.mul(self, other, context)

    def to_string(self, context):
        return 'url(%s)' % Literal.to_string(self, context)


class Var(Expr):

    def __init__(self, name, lineno=None):
        self.name = name
        self.lineno = lineno

    def evaluate(self, context):
        if self.name not in context:
            raise EvalException(self.lineno, 'variable %s is not defined' %
                                (self.name,))
        val = context[self.name]
        context[self.name] = FailingVar(self, self.lineno)
        try:
            return val.evaluate(context)
        finally:
            context[self.name] = val


class FailingVar(Expr):

    def __init__(self, var, lineno=None):
        Expr.__init__(self, lineno or var.lineno)
        self.var = var

    def evaluate(self, context):
        raise EvalException(self.lineno, 'Circular variable dependencies '
                            'detected when resolving %s.' % (self.var.name,))


class List(Expr):
    name = 'list'

    methods = {
        'length':   lambda x, c: Number(len(x.items)),
        'join':     lambda x, c, d=String(' '): String(d.value.join(
                                 a.to_string(c) for a in x.items))
    }

    def __init__(self, items, lineno=None):
        Expr.__init__(self, lineno)
        self.items = items

    def add(self, other):
        if isinstance(other, List):
            return List(self.items + other.items, lineno=self.lineno)
        return List(self.items + [other], lineno=self.lineno)

    def to_string(self, context):
        return u', '.join(x.to_string(context) for x in self.items)


class Parser(object):
    """
    Class with a bunch of methods that implement a tokenizer and parser.  In
    fact this class has two parsers.  One that splits up the code line by
    line and keeps track of indentions, and a second one for expressions in
    the value parts.
    """

    def preparse(self, source):
        """
        Do the line wise parsing and resolve indents.
        """
        rule = (None, [], [])
        vars = {}
        indention_stack = [0]
        state_stack = ['root']
        group_block_stack = []
        rule_stack = [rule]
        root_rules = rule[1]
        new_state = None

        lineiter = LineIterator(source, emit_endmarker=True)

        def fail(msg):
            raise ParserError(lineno, msg)

        def parse_definition():
            m = _def_re.search(line)
            if m is None:
                fail('invalid syntax for style definition')
            keys = tuple(x.strip() for x in m.group(1).split(','))
            return lineiter.lineno, keys, m.group(2)

        for lineno, line in lineiter:
            raw_line = line.rstrip().expandtabs()
            line = raw_line.lstrip()
            indention = len(raw_line) - len(line)

            # indenting
            if indention > indention_stack[-1]:
                if not new_state:
                    fail('unexpected indent')
                state_stack.append(new_state)
                indention_stack.append(indention)
                new_state = None

            # dedenting
            elif indention < indention_stack[-1]:
                for level in indention_stack:
                    if level == indention:
                        while indention_stack[-1] != level:
                            if state_stack[-1] == 'rule':
                                rule = rule_stack.pop()
                            elif state_stack[-1] == 'group_block':
                                name, part_defs = group_block_stack.pop()
                                for lineno, keys, val in part_defs:
                                    for key in keys:
                                        rule[2].append((lineno, name + '-' +
                                                        key, val))
                            indention_stack.pop()
                            state_stack.pop()
                        break
                else:
                    fail('invalid dedent')

            # new state but no indention. bummer
            elif new_state:
                fail('expected definitions, found nothing')

            # end of data
            if line == '__END__':
                break

            # root and rules
            elif state_stack[-1] in ('rule', 'root'):
                # new rule blocks
                if line.endswith(':'):
                    s_rule = line[:-1].rstrip()
                    if not s_rule:
                        fail('empty rule')
                    new_state = 'rule'
                    new_rule = (s_rule, [], [])
                    rule[1].append(new_rule)
                    rule_stack.append(rule)
                    rule = new_rule

                # if we in a root block we don't consume group blocks
                # or style definitions but variable defs
                elif state_stack[-1] == 'root':
                    if '=' in line:
                        m = _var_def_re.search(line)
                        if m is None:
                            fail('invalid syntax')
                        key = m.group(1)
                        if key in vars:
                            fail('variable "%s" defined twice' % key)
                        vars[key] = (lineiter.lineno, m.group(2))
                    else:
                        fail('Style definitions or group blocks are only '
                             'allowed inside a rule or group block.')

                # definition group blocks
                elif line.endswith('->'):
                    group_prefix = line[:-2].rstrip()
                    if not group_prefix:
                        fail('no group prefix defined')
                    new_state = 'group_block'
                    group_block_stack.append((group_prefix, []))

                # otherwise parse a style definition.
                else:
                    lineno, keys, value = parse_definition()
                    for key in keys:
                        rule[2].append((lineno, key, value))

            # group blocks
            elif state_stack[-1] == 'group_block':
                group_block_stack[-1][1].append(parse_definition())

            # something unparseable happened
            else:
                fail('unexpected character %s' % line[0])

        return root_rules, vars

    def parse(self, source):
        """
        Create a flat structure and parse inline expressions.
        """
        def handle_rule(rule, children, defs):
            def recurse():
                if defs:
                    result.append((get_selectors(), [
                        (k, self.parse_expr(lineno, v)) for
                        lineno, k, v in defs
                    ]))
                for child in children:
                    handle_rule(*child)

            local_rules = []
            reference_rules = []
            for r in rule.split(','):
                r = r.strip()
                if '&' in r:
                    reference_rules.append(r)
                else:
                    local_rules.append(r)

            if local_rules:
                stack.append(local_rules)
                recurse()
                stack.pop()

            if reference_rules:
                if stack:
                    parent_rules = stack.pop()
                    push_back = True
                else:
                    parent_rules = ['*']
                    push_back = False
                virtual_rules = []
                for parent_rule in parent_rules:
                    for tmpl in reference_rules:
                        virtual_rules.append(tmpl.replace('&', parent_rule))
                stack.append(virtual_rules)
                recurse()
                stack.pop()
                if push_back:
                    stack.append(parent_rules)

        def get_selectors():
            branches = [()]
            for level in stack:
                new_branches = []
                for rule in level:
                    for item in branches:
                        new_branches.append(item + (rule,))
                branches = new_branches
            return [' '.join(branch) for branch in branches]

        root_rules, vars = self.preparse(source)
        result = []
        stack = []
        for rule in root_rules:
            handle_rule(*rule)

        real_vars = {}
        for name, args in vars.iteritems():
            real_vars[name] = self.parse_expr(*args)

        return result, real_vars

    def parse_expr(self, lineno, s):
        def parse():
            pos = 0
            end = len(s)

            def process(token, group=0):
                return lambda m: (m.group(group), token)

            def process_string(m):
                value = m.group(0)
                try:
                    if value[:1] == value[-1:] and value[0] in '"\'':
                        value = value[1:-1].encode('utf-8') \
                                           .decode('string-escape') \
                                           .encode('utf-8')
                    elif value == 'rgb':
                        return None, 'rgb'
                    elif value in _colors:
                        return value, 'color'
                except UnicodeError:
                    raise ParserError(lineno, 'invalid string escape')
                return value, 'string'

            rules = ((_webkit_re, process('webkit', 1)),
                     (_operator_re, process('op')),
                     (_call_re, process('call', 1)),
                     (_value_re, lambda m: (m.groups(), 'value')),
                     (_color_re, process('color')),
                     (_number_re, process('number')),
                     (_rgba_re, process('rgba', 1)),
                     (_url_re, process('url', 1)),
                     (_string_re, process_string),
                     (_var_re, lambda m: (m.group(1) or m.group(2), 'var')),
                     (_whitespace_re, None))

            while pos < end:
                for rule, processor in rules:
                    m = rule.match(s, pos)
                    if m is not None:
                        if processor is not None:
                            yield processor(m)
                        pos = m.end()
                        break
                else:
                    raise ParserError(lineno, 'Syntax error')

        s = s.rstrip(';')
        return self.expr(TokenStream(lineno, parse()))

    def expr(self, stream, ignore_comma=False):
        args = [self.concat(stream)]
        list_delim = [(';', 'op')]
        if not ignore_comma:
            list_delim.append((',', 'op'))
        while stream.current in list_delim:
            stream.next()
            args.append(self.concat(stream))
        if len(args) == 1:
            return args[0]
        return List(args, lineno=stream.lineno)

    def concat(self, stream):
        args = [self.add(stream)]
        while stream.current[1] != 'eof' and \
              stream.current not in ((',', 'op'), (';', 'op'),
                                     (')', 'op')):
            args.append(self.add(stream))
        if len(args) == 1:
            node = args[0]
        else:
            node = ImplicitConcat(args, lineno=stream.lineno)
        return node

    def add(self, stream):
        left = self.sub(stream)
        while stream.current == ('+', 'op'):
            stream.next()
            left = Add(left, self.sub(stream), lineno=stream.lineno)
        return left

    def sub(self, stream):
        left = self.mul(stream)
        while stream.current == ('-', 'op'):
            stream.next()
            left = Sub(left, self.mul(stream), lineno=stream.lineno)
        return left

    def mul(self, stream):
        left = self.div(stream)
        while stream.current == ('*', 'op'):
            stream.next()
            left = Mul(left, self.div(stream), lineno=stream.lineno)
        return left

    def div(self, stream):
        left = self.mod(stream)
        while stream.current == ('/', 'op'):
            stream.next()
            left = Div(left, self.mod(stream), lineno=stream.lineno)
        return left

    def mod(self, stream):
        left = self.neg(stream)
        while stream.current == ('%', 'op'):
            stream.next()
            left = Mod(left, self.neg(stream), lineno=stream.lineno)
        return left

    def neg(self, stream):
        if stream.current == ('-', 'op'):
            stream.next()
            return Neg(self.primary(stream), lineno=stream.lineno)
        return self.primary(stream)

    def primary(self, stream):
        value, token = stream.current
        if token == 'number':
            stream.next()
            node = Number(value, lineno=stream.lineno)
        elif token == 'value':
            stream.next()
            node = Value(lineno=stream.lineno, *value)
        elif token == 'color':
            stream.next()
            node = Color(value, lineno=stream.lineno)
        elif token == 'rgb':
            stream.next()
            if stream.current == ('(', 'op'):
                stream.next()
                args = []
                while len(args) < 3:
                    if args:
                        stream.expect(',', 'op')
                    args.append(self.expr(stream, True))
                stream.expect(')', 'op')
                return RGB(tuple(args), lineno=stream.lineno)
            else:
                node = String('rgb')
        elif token == 'string':
            stream.next()
            # XXX: unescape
            node = String(value, lineno=stream.lineno)
        elif token == 'webkit':
            stream.next()
            node = Webkit(value, lineno=stream.lineno)
        elif token == 'rgba':
            stream.next()
            node = RGBA(value, lineno=stream.lineno)
        elif token == 'url':
            stream.next()
            node = URL(value, lineno=stream.lineno)
        elif token == 'var':
            stream.next()
            node = Var(value, lineno=stream.lineno)
        elif token == 'op' and value == '(':
            stream.next()
            if stream.current == (')', 'op'):
                raise ParserError(stream.lineno, 'empty parentheses are '
                                  'not valid. If you want to use them as '
                                  'string you have to quote them.')
            node = self.expr(stream)
            stream.expect(')', 'op')
        else:
            if token == 'call':
                raise ParserError(stream.lineno, 'You cannot call standalone '
                                  'methods. If you wanted to use it as a '
                                  'string you have to quote it.')
            stream.next()
            node = String(value, lineno=stream.lineno)
        while stream.current[1] == 'call':
            node = self.call(stream, node)
        return node

    def call(self, stream, node):
        method, token = stream.current
        assert token == 'call'
        stream.next()
        args = []
        while stream.current != (')', 'op'):
            if args:
                stream.expect(',', 'op')
            args.append(self.expr(stream))
        stream.expect(')', 'op')
        return Call(node, method, args, lineno=stream.lineno)


def convert(source, context=None):
    """Convert a CleverCSS file into a normal stylesheet."""
    return Engine(source).to_css(context)


def main():
    """Entrypoint for the shell."""
    import sys

    # help!
    if '--help' in sys.argv:
        print 'usage: %s <file 1> ... <file n>' % sys.argv[0]
        print '  if called with some filenames it will read each file, cut of'
        print '  the extension and append a ".css" extension and save. If '
        print '  the target file has the same name as the source file it will'
        print '  abort, but if it overrides a file during this process it will'
        print '  continue. This is a desired functionality. To avoid that you'
        print '  must not give your source file a .css extension.'
        print
        print '  if you call it without arguments it will read from stdin and'
        print '  write the converted css to stdout.'
        print
        print '  called with the --eigen-test parameter it will evaluate the'
        print '  example from the module docstring.'
        print
        print '  to get a list of known color names call it with --list-colors'

    # version
    elif '--version' in sys.argv:
        print 'CleverCSS Version %s' % VERSION
        print 'Licensed under the BSD license.'
        print '(c) Copyright 2007 by Armin Ronacher and Georg Brandl.'

    # evaluate the example from the docstring.
    elif '--eigen-test' in sys.argv:
        print convert('\n'.join(l[8:].rstrip() for l in
                      re.compile(r'Example::\n(.*?)__END__(?ms)')
                        .search(__doc__).group(1).splitlines()))

    # color list
    elif '--list-colors' in sys.argv:
        print '%s known colors:' % len(_colors)
        for color in sorted(_colors.items()):
            print '  %-30s%s' % color

    # read from stdin and write to stdout
    elif len(sys.argv) == 1:
        try:
            print convert(sys.stdin.read())
        except (ParserError, EvalException), e:
            sys.stderr.write('Error: %s\n' % e)
            sys.exit(1)

    # convert some files
    else:
        for fn in sys.argv[1:]:
            target = fn.rsplit('.', 1)[-1] + '.css'
            if fn == target:
                sys.stderr.write('Error: same name for source and target file'
                                 ' "%s".' % fn)
                sys.exit(2)
            src = file(fn)
            try:
                try:
                    converted = convert(src.read())
                except (ParserError, EvalException), e:
                    sys.stderr.write('Error in file %s: %s\n' % (fn, e))
                    sys.exit(1)
                dst = file(target, 'w')
                try:
                    dst.write(converted)
                finally:
                    dst.close()
            finally:
                src.close()


if __name__ == '__main__':
    main()



from django.template import Template


def add_origin(func):
    """
    We utilize the origin a lot internally for Django Skylark and when a template
    is created with one, we make it available off the Template instance itself.
    This allows us to continue to use the Django loader internals with minimal
    fuss and as much forward compatibility as possible

    Once this patch is applied, and it functions essentially like a decorator,
    we can acces the origin like this::

        >>> template = loader.get_template('some/template.html')
        >>> template.origin
        <django.template.loader.LoaderOrigin object at 0x102e1b710>
        >>> template.origin.name
        u'/Full/Path/to/some/template.html'
    """
    def wrapper(self, template_string,
                origin=None, name='<Unknown Template>'):
        func(self, template_string, origin, name)
        self.origin = origin
    return wrapper

setattr(Template, '__init__', add_origin(Template.__init__))



import py.test

from nose.tools import with_setup
from nose.plugins.attrib import attr
from nose.plugins.skip import SkipTest

from skylark import cssimgreplace

RAW = """
/* Common use case */
.class {
    background: white url("../images/img1.gif");
}
.class {
    background: white url('../images/img1.gif');
}
.class {
    background: white url(../images/img1.gif);
}

/* What about just an image in the same directory */
.class {
    background: white url("img1.gif");
}

/* And in the same directory, just phrased a bit differently */
.class {
    background: white url("./img1.gif");
}

/* Out a couple of directories */
.class {
    background: white url("../../../page2/images/img1.gif");
}

/* What about escaped characters */
.class {
    background: white url("../../../crazy\\\\page/images/img1.gif");
}

/* And the rest of these should not be touched */
.class {
    background: white url("http://google.com/img/logo.gif");
}
.class {
    background: white url("/cfcache/out/base/media/img/test.gif");
}
"""

EXPECTED = """
/* Common use case */
.class {
    background: white url("http://testserver/media/cfcache/out/dummyapp/page/media/images/img1.gif");
}
.class {
    background: white url('http://testserver/media/cfcache/out/dummyapp/page/media/images/img1.gif');
}
.class {
    background: white url(http://testserver/media/cfcache/out/dummyapp/page/media/images/img1.gif);
}

/* What about just an image in the same directory */
.class {
    background: white url("http://testserver/media/cfcache/out/dummyapp/page/media/css/img1.gif");
}

/* And in the same directory, just phrased a bit differently */
.class {
    background: white url("http://testserver/media/cfcache/out/dummyapp/page/media/css/img1.gif");
}

/* Out a couple of directories */
.class {
    background: white url("http://testserver/media/cfcache/out/dummyapp/page2/images/img1.gif");
}

/* What about escaped characters */
.class {
    background: white url("http://testserver/media/cfcache/out/dummyapp/crazy\\\\page/images/img1.gif");
}

/* And the rest of these should not be touched */
.class {
    background: white url("http://google.com/img/logo.gif");
}
.class {
    background: white url("/cfcache/out/base/media/img/test.gif");
}
"""

def test_will_replace_url_values():
    result = cssimgreplace.relative_replace(RAW,
        'dummyapp/page/media/css/',
        'http://testserver/media/cfcache/out/')
    assert result == EXPECTED



import copy
import os
import yaml

from django.template import Template, loader

from skylark import plans
from skylark.conf import settings
from skylark import chirp


class Renderer(object):
    """
    The base class that performs the heavy lifting of taking page instructions
    and rendering them into actual HTML.  This class is not meant to be used by
    itself but instead extended.  The template_name variables should contain
    the template that you wish to use to render whatever kind of page you need.
    In other words, check out Xhtml401Transitional for an example.
    """

    # Used to place a doctype at the beginning of a rendered page.  This only
    # applies to html and xhtml pages that are rendered with a PageAssembly.
    doctype = ('<!DOCTYPE html>')

    # What template do we use to render?
    template_name = 'skylark/html5.html'
    snippet_template_name = 'skylark/htmlsnippet.html'

    def __init__(self, page_instructions, context, render_full_page=True,
                 omit_media=False):
        self.page_instructions = page_instructions
        self.context = context
        self.render_full_page = render_full_page
        self.omit_media = omit_media

    def render(self):
        """
        Takes a chunk of page instructions and renders a page according to the
        rules found within

        This return a string representing the HTML or similar output
        """
        assert self.page_instructions.body, \
            'The body has not been specified in the page instructions ' + \
            '(body: in your yaml file)'

        if self.render_full_page:
            assert self.page_instructions.title, \
                'The title has not been specified in the page ' + \
                'instructions (title: in your yaml file)'

        plan = plans.get_for_context(self.context,
            self.page_instructions.render_full_page)
        prepared_instructions = plan.prepare(self.page_instructions,
            omit_media=self.omit_media)

        render_context = copy.copy(self.context)
        render_context['cache_url'] = settings.SKYLARK_CACHE_URL
        render_context['doctype'] = self.doctype
        render_context['prepared_instructions'] = prepared_instructions
        render_context['is_instrumented'] = chirp.is_instrumented()

        if self.render_full_page:
            t = self.template_name
        else:
            t = self.snippet_template_name
        self.template = loader.get_template(t)

        return self.template.render(render_context)



import sys
import py.test

from nose.tools import with_setup
from nose.plugins.attrib import attr
from nose.plugins.skip import SkipTest
from skylark import *
from skylark.assembly import *
from skylark.page import PageAssembly
from skylark.snippet import SnippetAssembly
from django.conf import settings

from skylark.tests import *
from skylark import chirp


def teardown_chirp():
    teardown()
    chirp.instrument_site(False)
    chirp.test_registry.clear()


@with_setup(setup, teardown_chirp)
def test_instrument():
    assert not chirp.is_instrumented()
    chirp.instrument_site(True)
    py.test.raises(chirp.ChirpError, chirp.instrument_site, True)
    assert chirp.is_instrumented()
    chirp.instrument_site(False)
    assert not chirp.is_instrumented()
    settings.SKYLARK_CHIRP_INSTRUMENTED = True
    # This only get's triggered if we render a page, so let's do that
    request = get_request_fixture()
    c = RequestContext(request, {})
    sa = SnippetAssembly('dummyapp/snippet/snippet.yaml', c)
    content = sa.dumps()
    assert chirp.is_instrumented()
    settings.SKYLARK_CHIRP_INSTRUMENTED = False


@with_setup(setup, teardown_chirp)
def test_snippet_render():
    request = get_request_fixture()
    c = RequestContext(request, {})
    sa = SnippetAssembly('dummyapp/snippet/snippet.yaml', c)

    content = sa.dumps()

    assert not '<html' in content
    assert not '<head' in content
    assert not '<body' in content
    assert not '<link' in content

    assert "dojo.registerModulePath('DynamicApp.Snippet'" in content

    exist(
        'out/dynamicapp/media/js/chirploaded.js',
        'out/dynamicapp/media/js/templates/chirploaded.html',
    )

    assert 'This is my snippet test' in content
                              

@with_setup(setup, teardown_chirp)
def test_chirp_renders_in_page():
    request = get_request_fixture()
    c = RequestContext(request, {})
    pa = PageAssembly('dummyapp/page/chirp.yaml', c)

    content = pa.dumps()

    exist(
        'addon/dojo/dojo.js',
        'addon/dojox/timing/_base.js',
        'out/dummyapp/page/media/img/notreferenced.png',
        'out/dummyapp/page/media/img/test.png',
        'out/dummyapp/page/media/js/sample.js',
        'out/dummyapp/page/media/js/templates/notreferenced.html',
        'out/dummyapp/page/media/js/templates/sample.js',
        'out/dummyapp/tag/media/css/screen.css',
        'out/dummyapp/tag/media/img/testimage.png',
        'out/dummyapp/tag/media/js/templates/test.html',
        'out/dynamicapp/media/js/chirploaded.js',
        'out/dynamicapp/media/js/templates/chirploaded.html',
    )

    assert "dojo.registerModulePath('DynamicApp.Page'" in content
    assert "dojo.require('DynamicApp.Page.Controller');" in content
    assert "dojo.require('DynamicApp.Page.View');" in content
    assert 'dummyapp/tag/media/css/screen.css' in content
    assert 'dummyapp/page/media/js/sample.js' in content
    assert 'media/cfcache/out/dynamicapp' in content
    assert 'addon/dojo/dojo.js' in content


@with_setup(setup, teardown_chirp)
def test_chirp_dojo_settings():
    settings.SKYLARK_DOJO_VIA_CDN_AOL = True

    request = get_request_fixture()
    c = RequestContext(request, {})
    pa = PageAssembly('dummyapp/page/chirp.yaml', c)
    content = pa.dumps()
    assert 'http://o.aolcdn.com/dojo/1.4/dojo/dojo.xd.js' in content

    settings.SKYLARK_DOJO_VIA_CDN_AOL = None

    settings.SKYLARK_DOJO_VIA_URL = 'http://testdojo.com/dojo.js'

    request = get_request_fixture()
    c = RequestContext(request, {})
    pa = PageAssembly('dummyapp/page/chirp.yaml', c)
    content = pa.dumps()
    assert 'http://testdojo.com/dojo.js' in content

    settings.SKYLARK_DOJO_VIA_URL = None


@with_setup(setup, teardown_chirp)
def test_chirp_test_registry_add():
    py.test.raises(chirp.ChirpError, chirp.test_registry.add, None)

    chirp.test_registry.add('/someurl', name='Some Test')

    names = [ i.name for i in chirp.test_registry.list() ]
    urls = [ i.url for i in chirp.test_registry.list() ]

    assert '/someurl' in chirp.test_registry
    assert '/nothere' not in chirp.test_registry
    assert '/someurl' in urls
    assert 'Some Test' in names

    chirp.test_registry.add('/another', name='A special test')

    assert len(chirp.test_registry) == 2


@with_setup(setup, teardown_chirp)
def test_chirp_autodiscover():
    chirp.autodiscover()
    assert 'dummyapp.chirp' in chirp._chirp_modules

    chirp.instrument_site(True)
    assert 'dummyapp.chirp' in sys.modules


@with_setup(setup, teardown_chirp)
def test_chirp_testcase_is_included():
    assert not chirp.is_instrumented()
    chirp.instrument_site(True)

    request = get_request_fixture()
    c = RequestContext(request, {})
    pa = PageAssembly('dummyapp/page/chirp.yaml', c)
    content = pa.dumps()

    assert 'dojo.require(\'ChirpTools.TestRunner' in content
    assert 'dojo.require(\'ChirpTools.Mvc' in content
    assert 'ChirpTools.TestRunner.TestCaseCollector' in content


@with_setup(setup, teardown_chirp)
def test_chirp_includes_tests():
    assert not chirp.is_instrumented()
    chirp.instrument_site(True)

    request = get_request_fixture()
    c = RequestContext(request, {})
    pa = PageAssembly('dummyapp/page/chirp.yaml', c)
    content = pa.dumps()

    assert "dojo.require('DynamicApp.Page.Test')" in content

    chirp.instrument_site(False)

    c = RequestContext(request, {})
    pa = PageAssembly('dummyapp/page/chirp.yaml', c)
    content = pa.dumps()

    assert "dojo.require('DynamicApp.Page.Test')" not in content



import copy
import os
import re
import filecmp
import shutil
import hashlib
import pickle
import subprocess
from urlparse import urljoin

from django.template import Template, TemplateDoesNotExist
from django.template import loader
from django.template.loaders import filesystem
from django.template.loaders import app_directories
from django.utils.functional import memoize

from skylark.conf import settings
from skylark.processor import clevercss
from skylark import chirp
from skylark import cssimgreplace


class BadOption(Exception):
    """
    If a plan option is being specified that is not supported.
    """
    pass


class BadPlanSituation(Exception):
    """
    Used to indicate that some condition is causing the plan to fail.  These
    typically cannot be resolved until the developer changes some condition to
    fix it.
    """
    pass


class DojoModuleResolution(Exception):
    """
    We extract "dojo.require" method calls from the Javascript source code and
    try to find the files they refer to.  If for one reason or another we
    cannot find it, this is raised.
    """
    pass


class SkipDojoModule(Exception):
    """
    Used to indicate the module should not be located and parsed for other dojo
    requirements.
    """
    pass


def find_directory_from_loader(page_instructions, asset):
    from django.template.loaders.app_directories import app_template_dirs
    template_dirs = list(settings.TEMPLATE_DIRS) + list(app_template_dirs)

    for dir in template_dirs:
        asset_dir = os.path.join(dir, asset)
        if os.path.isdir(asset_dir):
            return asset_dir

    raise TemplateDoesNotExist('Unable to find a directory within known '
        'template directories: %s' % asset)


def process_clevercss(source, filepath):
    """
    This is part of the processing_funcs that Renderer will use to perform any
    special transformations or filtering on the output of a rendered template.

    This particular one uses CleverCSS to process a meta-css file and convert
    it into normal css.  More info at http://sandbox.pocoo.org/clevercss/
    """
    return clevercss.convert(source)


def process_lessjs(source, filepath):
    """
    Less is a CSS processor, that extends the syntax and adds variables,
    arithmetic, and other goodies

    Less.js author: http://cloudhead.io/
    GitHub repo: http://github.com/cloudhead/less.js
    """
    # This is a simple pass through, we don't need to do anything for less.js
    # to work
    return source


def process_lessc(source, filepath):
    """
    Less is a CSS processor, that extends the syntax and adds variables,
    arithmetic, and other goodies. This process precompiles the Less source
    into proper CSS.

    Less.js author: http://cloudhead.io/
    GitHub repo: http://github.com/cloudhead/less.js
    """
    return subprocess.check_output(['lessc', filepath])


class BasePlan(object):
    """
    Base class that all the plans can subclass.  It provides common things like
    copying directories, and other fancy pants things like that.
    """

    cache_prefix = 'out'

    __media_source_cache = {}

    options = {
        'minify_javascript': True,
    }

    """
    These are a set of special functions that can be used to manipulate the
    source of the page.  The way these get triggered is through the attribute
    "process:" in the yaml file.  Here's a quick example.

        static: myapp/common.css
        process: clevercss

    As the instruction is parsed, we come across process and if it matches a
    defined processing_funcs we will provide the source of the page and allow
    it to be modified by the function.
    """
    processing_funcs = {
        'clevercss': process_clevercss,
        'lessjs': process_lessjs,
        'lessc': process_lessc}

    make_css_urls_absolute = False

    def __init__(self, context, render_full_page):
        self._prepare_assets_cache = {}
        setattr(self, '_prepare_assets', memoize(
            self._prepare_assets, self._prepare_assets_cache, 2))

        self.cache_root = os.path.join(
            settings.SKYLARK_CACHE_ROOT, self.cache_prefix)
        self.cache_url = urljoin(settings.SKYLARK_CACHE_URL,
            '%s/' % self.cache_prefix)
        self.context = context
        self.render_full_page = render_full_page

        """
        As we process the page instructions, we gather the output we need to
        convert this into an html page inside this dictionary
        """
        self.prepared_instructions = {
            'meta': [],
            'js': [],
            'css': [],
            'chirp': [],
        }

        self.prepared_instructions['render_full_page'] = self.render_full_page
        self.prepared_instructions['cache_prefix'] = '%s/' % self.cache_prefix

        if not os.path.exists(self.cache_root):
            os.makedirs(self.cache_root)

    @classmethod
    def set_options(*args, **kwargs):
        """
        Various options for how the plans behave
        """
        for option in kwargs:
            if option not in BasePlan.options:
                raise BadOption('%s is not a valid, must be a combination '
                    'of %s' % (option, ','.join(BasePlan.options.keys(),)))
        BasePlan.options.update(kwargs)

    def _get_media_stat(self, template_name):
        """
        Performs a os.stat on template_name, raising TemplatePathDoesNotExist
        if the template_name is not file based.
        """
        source, filepath = self._get_source_filepath(template_name)
        return os.stat(filepath)

    def _get_source_filepath(self, template_name):
        """
        Utilizes some of Django's internals to retrive the source code and
        filepath for a template while bypassing the normal compile behavior
        """
        # FIXME: we are using the Django file system loader here, this API
        # might change in the future.  Should we isolate this into a
        # functional programming style section instead of deep in this
        # class?
        try:
            source, filepath = \
                filesystem._loader.load_template_source(template_name)
        except TemplateDoesNotExist:
            source, filepath = \
                app_directories._loader.load_template_source(template_name)
        return source, filepath

    def _get_media_source(self, template_name, process_func=None,
        context=None, no_render=False):
        """
        Responsible for taking a template and generating the contents.

            * Renders the template with the given context if applicable
            * Passes it through the process function if provided
        """
        cache = self.__media_source_cache

        mem_args = (template_name, process_func)
        if mem_args in cache and not context and not settings.DEBUG:
            source = cache[mem_args]
            is_cached = True
        else:
            if context and not no_render:
                template = loader.get_template(template_name)
                return template.render(context), False

            source, filepath = self._get_source_filepath(template_name)

            if process_func:
                source = process_func(source, filepath)

            cache[mem_args] = source
            is_cached = False

        return source, is_cached

    def _copy_to_media(self, template_name, source=''):
        """
        Part of our goal here is to make the placement of media a transparent
        deal.  Django does not currently make this easy, you typically have to
        handle your media in a pretty manual fashion.

        This method takes a file that is somewhere in your template path (for
        example /blog/templates/blog/list/media/css/screen.css and copies it to
        the cache.  It ends up having the same directory structure, so in the
        end gets copied to MEDIA_ROOT/cfcache/blog/media/css/screen.css.
        """
        dirpath = os.path.join(self.cache_root, os.path.dirname(template_name))
        filename = os.path.basename(template_name)
        fullpath = os.path.join(dirpath, filename)

        if not os.path.isfile(fullpath) or settings.DEBUG:
            if not os.path.exists(dirpath):
                os.makedirs(dirpath)

            f = open(fullpath, 'w')
            f.write(source)
            f.close()

        return urljoin(self.cache_url, template_name), filename

    def _get_processing_function(self, process_func):
        """
        Retrieves one of the processing functions that can transform our source
        into something else

        An example here is using CleverCSS:

            css:
                - static: screen.css
                  process: clevercss
        """
        if not process_func:
            return None

        if process_func in self.processing_funcs:
            return self.processing_funcs[process_func]
        else:
            raise AttributeError('Could not find a process function matching '
                '%s, available ones are: %s' %
                (process_func, ', '.join(self.processing_funcs.keys()),))

    def _fix_css_urls(self, page_instruction, css_source):
        """
        When we are rolling CSS files up the are placed in the root of the
        cache directory.  This breaks url values like url(../img/logo.gif).

        To fix this, we need to manipulate the url values and provide a
        corrected URL.
        """
        if 'static' in page_instruction:
            template_name = page_instruction['static']
        elif 'inline' in page_instruction:
            template_name = page_instruction['inline']

        return cssimgreplace.relative_replace(
            css_source,
            os.path.dirname(template_name),
            self.cache_url)

    def _prepare_file(self, item_name, page_instructions):
        """
        This method takes lines from our YAML files and decides what to do with
        them.

        An example is::

            js:
                - url: http://somesite.com/somefile.html

            css:
                - static: blog/index/

        This method is a factored out version of prepare_css and prepare_js.
        """
        if item_name not in self.prepared_instructions:
            self.prepared_instructions[item_name] = []

        for instruction in getattr(page_instructions, item_name):
            item = copy.copy(instruction)

            if 'url' in instruction:
                item['location'] = instruction['url']

            else:
                template_name = context = process_func = None

                if 'process' in instruction:
                    process_func = self._get_processing_function(
                        instruction.get('process'))

                template_name = instruction.get('static', False) or \
                    instruction.get('inline', False)

                assert template_name, (
                    'You must provide either "static" or "inline" properties '
                    'that point to a file, provided object was %r'
                    % instruction)

                if 'inline' in instruction:
                    context = self.context
                else:
                    context = None

                source, is_cached = self._get_media_source(
                    template_name, process_func, context)

                if 'css' in item_name and self.make_css_urls_absolute \
                   and not is_cached:
                    source = self._fix_css_urls(instruction, source)

                if 'static' in instruction:
                    location, filename = self._copy_to_media(
                        template_name, source)
                    item['location'] = location
                elif 'inline' in instruction:
                    item['source'] = source

                if 'include' in instruction and not \
                   instruction['include']:
                    if 'inline' in instruction:
                        raise AttributeError('You have specified inline and '
                            'include: false, these really don\'t make sense '
                            'together')
                    continue

            self.prepared_instructions[item_name].append(item)

    def _prepare_assets(self, page_instructions, assets=None):
        """
        There are some special cases when working with css and javascript that
        we make allowances for.

        The first is css.  When you author your css it's normal to see
        references to images like this::

            background-image: url(../img/header/background.png)

        CSS authors are used to referencing images this way.  Since we cache
        css files from within the app to the MEDIA_ROOT, we need to also copy
        images that may be used.  We do this by looking for media/img relative
        to the yaml file that was used to generate the page instructions and
        copy this entire directory to the cache.

        The same thing goes for Javascript templates.  They are not a widely
        used item, but we've included them because it's part of what our
        original goal was when developing this app.

        You can put HTML files in media/js/templates and the entire templates
        directory will be copied into the MEDIA_ROOT in the appropriate spot.
        This way your javascript files can utilize them without having to worry
        about where they are.

        This method will work with directories that are relative to the YAML
        file or the app's templates directory.  The following will essentially
        copy the same directory to the cache:

            self._prepare_assets(pi, ('media/js',))
            self._prepare_assets(pi, ('blog/list/media/js',))

        """
        assert type(assets) == tuple or type(assets) == list

        for yaml in page_instructions.yaml:
            # yaml = app/page/page.yaml
            template, origin = loader.find_template(yaml)
            filepath = template.origin.name

            # /Users/me/Development/app/templates/app/page/page.yaml
            yaml_basedir = os.path.dirname(yaml)
            # app/page
            template_basedir = filepath[:filepath.find(yaml)]
            # /Users/me/Development/app/templates

            for asset in assets:
                # directory = /media/js/templates
                if not yaml_basedir in asset:
                    # The user might be specifying the directory relative to
                    # the yaml file itself, so we'll add it for them if they
                    # gave us something like 'media/js/templates'
                    directory = os.path.join(yaml_basedir, asset)
                else:
                    directory = asset

                sourcedirectory = os.path.join(template_basedir, directory)

                if not os.path.isdir(sourcedirectory):
                    # We're going to try and find it somewhere else, it may not
                    # be relative to the YAML file
                    #
                    # This is quite possible if the yaml file is processing a
                    # "chirp:" attribute.
                    try:
                        sourcedirectory = find_directory_from_loader(
                            page_instructions, asset)
                        # We need to reset this, it has the yaml_basedir on it
                        # at this point
                        directory = asset
                    except TemplateDoesNotExist:
                        continue

                if not os.path.isdir(sourcedirectory):
                    continue

                cachedirectory = os.path.join(self.cache_root, directory)

                if os.path.isdir(cachedirectory):
                    if self._assets_are_stale(sourcedirectory, cachedirectory):
                        shutil.rmtree(cachedirectory)
                    else:
                        continue

                shutil.copytree(sourcedirectory, cachedirectory)

                if settings.FILE_UPLOAD_PERMISSIONS is not None:
                    for root, dirs, files in os.walk(cachedirectory):
                        for momo in files:
                            os.chmod(os.path.join(root, momo),
                                settings.FILE_UPLOAD_PERMISSIONS)
                        for momo in dirs:
                            os.chmod(os.path.join(root, momo), 02750)

    def _assets_are_stale(self, sourcedirectory, cachedirectory):
        """
        Looks through the given directories, determining if they are different
        """
        comparison = filecmp.dircmp(sourcedirectory, cachedirectory, [], [])
        if comparison.left_only or comparison.right_only:
            # We have files in one directory and not the other
            return True
        if comparison.diff_files:
            # Some of the files have changed
            return True

        return False

    def prepare_title(self, page_instructions):
        """
        Prepares the title for the page
        """
        template = Template(str(page_instructions.title))
        self.prepared_instructions['title'] = unicode(
            template.render(self.context))

    def prepare_body(self, page_instructions):
        """
        Takes the body section and renders it, storing it in
        prepared_instructions
        """
        template = loader.get_template(str(page_instructions.body))
        self.prepared_instructions['body'] = unicode(
            template.render(self.context))

    def prepare_meta(self, page_instructions):
        """
        Prepares the meta section
        """
        self.prepared_instructions['meta'] = page_instructions.meta

    def prepare_chirp(self, page_instructions):
        chirp_instructions = page_instructions.chirp

        self.prepared_instructions['chirp'] = chirp_instructions

        for chirp_module in chirp_instructions:
            assert 'namespace' in chirp_module, ('You are missing the '
                'namespace attribute for this item')
            assert 'location' in chirp_module, ('You are missing the '
                 'location attribute for this item')
            assert 'require' in chirp_module or 'tests' in chirp_module, (
                 'You are missing the require list for this item')

            if not 'require' in chirp_module:
                chirp_module['require'] = []
            chirp_module['needs_registration'] = True

            namespace = chirp_module['namespace']
            location = chirp_module['location']
            require = chirp_module['require']

            tests = []
            if chirp.is_instrumented():
                tests = chirp_module.get('tests', tests)

            require.extend(tests)

            """
            We're going to copy all the files that are in this directory to the
            cache.  This is not ideal, as not all the files may be used but the
            alternative is we ask the user specifically which ones they need.
            Since this is within the context of Dojo, that may not make the
            most sense.
            """
            self._prepare_assets(page_instructions, (location,))

    def prepare(self, page_instructions, omit_media=False):
        self.page_instructions = page_instructions

        self.prepare_title(page_instructions)
        self.prepare_body(page_instructions)
        self.prepare_meta(page_instructions)

        if not omit_media:
            """
            This can get flipped to True if we are nested inside another
            rendering.

            For example, if a snippet assembly is being used in a template tag
            and we are in the render function as part of that tag, there is
            going to be another page assembly responsible for actually
            rendering the page.  It will handle the media for us, so we clean
            out the media sections of our prepared instructions here to prevent
            duplication.
            """
            self.prepare_js(page_instructions)
            self.prepare_css(page_instructions)
            self.prepare_chirp(page_instructions)

        return self.prepared_instructions


class RollupPlan(object):
    __rollup_last_modifieds = {}

    """
    These are extra methods that are needed for rolling up files

    A rollup is when you take multiple files and concatenate them into one.
    """
    def _concat_files(self, instructions, fix_css_urls=False):
        source = []
        for i in instructions:
            processed = ''
            if 'source' in i:
                processed = i['source']
            else:
                process_func = self._get_processing_function(i.get('process'))
                processed, is_cached = self._get_media_source(
                    i['static'], process_func, no_render=True)
            if fix_css_urls:
                processed = self._fix_css_urls(i, processed)
            if isinstance(processed, basestring):
                source.append(processed)
            else:
                source.extend(processed)
        return "\n".join(source)

    def _make_filename(self, files):
        files.sort()
        return hashlib.md5('%s%s' % (settings.SKYLARK_PLANS_ROLLUP_SALT,
            pickle.dumps(files))).hexdigest()

    def _prepare_rollup(self, attr, rollup, keep, insert_point, **kwargs):
        if not keep and not rollup:
            return
        rollup_instruction = self._rollup_static_files(
            rollup, attr, kwargs.get('minifier', None),
            kwargs.get('wrap_source', None))
        if not rollup_instruction:
            return
        other_instruction = self.prepared_instructions[attr]

        self.prepared_instructions[attr] = \
            other_instruction[:insert_point] + \
            [rollup_instruction] + \
            other_instruction[insert_point:]

    def _instructions_have_lessjs(self, instructions):
        """
        Goes through the instructions and determine if any of the source files
        are set to be processed with lessjs.
        """
        for i in instructions:
            try:
                if i['process'] == 'lessjs':
                    return True
            except KeyError:
                # No process, but that's OK
                pass
        return False

    def _rollup_static_files(self, instructions, extension, minifier=None,
        wrap_source=None):
        """
        Creates one file from a list of others.  It also minifies the source
        using the appropriate function

        wrap_source is a tuple with a length of 2.  It can be used to prepend
        and append content to the rolled up file.  For example the following
        tuple would wrap the entire source in a dojo.addOnLoad().

        wrap_source = ('dojo.addOnLoad(function(){', '})',)
        """
        fix_css_urls = True if 'css' in extension else False
        is_lessjs = self._instructions_have_lessjs(instructions)

        def nop_minifier(arg):
            """
            A minifier that does nothing, but is callable
            """
            return arg

        # Figure out a name
        files = [i['static'] for i in instructions]

        basename = '%s.%s' % (self._make_filename(files), extension,)
        filename = os.path.join(self.cache_root, basename)
        location = urljoin(self.cache_url, basename)
        retval = {'location': location}

        if is_lessjs:
            retval['process'] = 'lessjs'

        if not files:
            return None

        lastmod = max([self._get_media_stat(i).st_mtime for i in files])

        if os.path.isfile(filename) and \
           filename in self.__rollup_last_modifieds and \
           self.__rollup_last_modifieds[filename] == lastmod:
            # Nothing has changed since we last saw this instruction set
            return retval
        self.__rollup_last_modifieds[filename] = lastmod

        if not wrap_source:
            wrap_source = ('', '',)

        if not minifier or is_lessjs:
            """
            If minifier is not defined we use a no-operate version

            If lessjs is used, we can't alter the original file because it will
            throw the parser off.  So we turn off the minification
            """
            minifier = nop_minifier

        if not os.path.isfile(filename) or settings.DEBUG:
            f = open(filename, 'w')
            source = minifier(self._concat_files(instructions, fix_css_urls))
            f.write('%s\n%s\n%s' % (wrap_source[0], source, wrap_source[1],))
            f.close()

        return retval

    def __dojo_register_module_path(self, namespace, basename):
        location = urljoin(self.cache_url, basename)
        js_tmp= "dojo.registerModulePath('%(namespace)s', '%(location)s');"
        return js_tmp % {'namespace': namespace, 'location': location}

    def _chirp_needs_registration(self, chirp_instructions, needs_registration):
        for chirp_module in chirp_instructions:
            chirp_module['needs_registration'] = needs_registration

    def _rollup_chirp(self, chirp_instructions):
        local_modules = []
        skip_modules = []

        for chirp_module in chirp_instructions:
            if 'require' not in chirp_module:
                continue
            namespace = chirp_module['namespace']
            location = chirp_module['location']
            require = chirp_module['require']

            # Figure out the path for each requirement based on namespace and
            # location
            for req in require:
                filename = '%s.js' % req.replace('%s.' % namespace, '')
                req_location = os.path.join(location, filename)
                source, is_cached = self._get_media_source(req_location)

                # Now we need to add a dojo.registerModulePath to this
                source = '%s\n' % self.__dojo_register_module_path(
                    namespace, location) + source
                # And tell the instructions it no longer needs to worry about
                # the registration of this module
                local_modules.append({'name': req, 'static': req_location,
                    'source': source})

        if len(skip_modules) > 0:
            # Well, we decided to skip some modules and registration of the
            # module still needs to be listed in the script block below the
            # rolled up files
            self._chirp_needs_registration(chirp_instructions, True)
        else:
            self._chirp_needs_registration(chirp_instructions, False)

        self._local_modules = local_modules
        self._skip_modules = skip_modules

        roll_modules = []

        for mod in self._local_modules:
            self._extract_dojo_requires(roll_modules, mod['name'],
                mod['static'], mod['source'])

        return roll_modules

    def _extract_dojo_requires(self, roll_modules, name, static, source):
        if name in [i['name'] for i in roll_modules]:
            return

        req_pattern = re.compile(
            r'\s*dojo\.require\((\'|")(?P<mod>[^\'"]+)(\'|")\)')

        matches = []
        lines = source.splitlines()
        for line in lines:
            match = req_pattern.match(line)
            if match:
                matches.append(match.groupdict()['mod'])

        for req_mod in matches:
            try:
                req_mod_path = self._resolve_dojo_module_path(req_mod)
                if not req_mod_path:
                    # We couldn't locate it
                    raise DojoModuleResolution(
                        'Could not resolve %s' % req_mod)
                try:
                    req_mod_source, is_cached = self._get_media_source(
                        req_mod_path, no_render=True)
                except TemplateDoesNotExist as tdne:
                    raise TemplateDoesNotExist('Trying to find %s at path %s '
                       'while processing dojo.requires for %s' % (
                            req_mod, req_mod_path, name))
            except SkipDojoModule as sdm:
                continue

            self._extract_dojo_requires(
                roll_modules, req_mod, req_mod_path, req_mod_source)

        # Filter this module out of the prepared instructions, we've now set it
        # to roll in with the other JS
        self._unprepare_chirp_dojo_module(name)

        roll_modules.append({'name': name, 'static': static, 'source': source})

    def _resolve_dojo_module_path(self, mod):
        if not settings.SKYLARK_DOJO_VIA_INTERNALBUILD:
            # We can only make this work if we use the internal build
            return None

        # Are we supposed to skip this one?
        skip = [i['static'] for i in self._skip_modules if i['name'] == mod]
        if skip:
            raise SkipDojoModule()

        # It could be in our local_modules
        local = [i['static'] for i in self._local_modules if i['name'] == mod]
        if local:
            return local[0]

        mod_parts = mod.split('.')
        if mod_parts[0].lower() != 'dojo' and mod_parts[0].lower() != 'dojox':
            # We have a problem Houston, this thing is not a Dojo module but
            # it's not included in the local modules
            namespaces = [i['namespace'] for i in self.page_instructions.chirp]
            raise DojoModuleResolution('While processing %s, dojo module %s '
               'was not found, the following namespaces are registered: %s' %
               (self.page_instructions.root_yaml, mod, ', '.join(namespaces)))
        # Perhaps it's a dojo or dojox module
        mod_path = '%s.js' % os.path.join('chirp', 'media', *mod_parts)

        return mod_path

    def _unprepare_chirp_dojo_module(self, mod):
        for chirp_instruction in self.prepared_instructions['chirp']:
            if 'require' not in chirp_instruction:
                continue
            if mod in chirp_instruction['require']:
                chirp_instruction['require'].remove(mod)
                return



from django.core.management.base import BaseCommand, CommandError
from django.core.cache import cache
from django.utils.importlib import import_module
from optparse import make_option
from messages import *
from skylark.conf import settings

import os
import hashlib


class Command(BaseCommand):
    help = MESSAGE_HELP

    option_list = BaseCommand.option_list + (
        make_option("--noconfirm", "-n", dest="noconfirm", action="store_true",
            help="Whether we show you a summary of what will happen and let "
                "you confirm"),
        make_option("--app", "-a", dest="app",
            help="The application to add a page to"),
        make_option("--page", "-p", dest="page",
            help="The page name you wish to create inside the app"),
    )

    def handle(self, *args, **options):
        if len(args) > 0:
            # Display the help and quit this foolishness
            self.print_help('skylarkpage', '')
            return

        # Here is our args
        app = options.get('app')
        page = options.get('page')
        noconfirm = True if options.get('noconfirm') == True else False
        self.confirm = not noconfirm

        if not app:
            raise CommandError("You must provide the app name, here's what "
               "you have in settings: %s" % settings.INSTALLED_APPS)

        if not page:
            raise CommandError("You must provide the page name")

        # We look for the application in the directory, and for a templates
        # directory
        try:
            appobject = import_module(app)
            directory = os.path.dirname(appobject.__file__)
        except ImportError:
            raise CommandError("The app %s could not be imported.  Is it on "
               "your python path?" % app)

        tmp_dir = os.path.join(directory, 'templates')

        if not os.path.isdir(tmp_dir) and self.confirm:
            create_tmp_dir = raw_input(self.style.NOTICE(
                MESSAGE_CONFIRM_TEMPLATE_CREATION % tmp_dir))
            if create_tmp_dir == 'yes':
                os.mkdir(tmp_dir)

        if not os.path.isdir(tmp_dir):
            raise CommandError("The templates directory is missing")

        page_dir = os.path.join(app.replace(".", os.sep), page)

        dirs_to_create = (
            os.path.join(tmp_dir, page_dir, 'media'),
            os.path.join(tmp_dir, page_dir, 'media', 'css'),
            os.path.join(tmp_dir, page_dir, 'media', 'img'),
            os.path.join(tmp_dir, page_dir, 'media', 'js'),
            os.path.join(tmp_dir, page_dir, 'media', 'js', 'templates'),
        )

        files_to_create = (
            os.path.join(tmp_dir, page_dir, 'media', 'css', "screen.css"),
            os.path.join(tmp_dir, page_dir, 'media', 'js', "%s.js" % page),
        )

        if self.confirm:
            ok_to_create = self.__confirm_creation(
                dirs_to_create, files_to_create)

            if ok_to_create:
                self.__create_these(dirs_to_create, files_to_create)
            else:
                raise CommandError("Ok, fine, we won't touch it if you've "
                   "changed your mind")
        else:
            self.__create_these(dirs_to_create, files_to_create)

        """
        Now we're gonna make the YAML and HTML files
        """
        context = {'app': app, 'page': page, 'page_dir': page_dir}
        yaml_file = os.path.join(tmp_dir, page_dir, "%s.yaml" % page)
        html_file = os.path.join(tmp_dir, page_dir, "%s.html" % page)

        self.__fill_file(yaml_file, CONTENT_YAML, context)
        self.__fill_file(html_file, CONTENT_HTML, context)

        print self.style.NOTICE(
            "\nYou can do something like this in your view now\n")

        print CONTENT_VIEW % {'app': app, 'page': page, 'page_dir': page_dir}

    def __fill_file(self, file, content_format, context={}):
        """
        Acts as a poor mans template engine, replacing the tokens in
        content_format with stuff in context and then writes it to
        file if it doesn't exist
        """
        if os.path.isfile(file):
            print self.style.ERROR(
                "%s already exists, we won't mess with it" % file)
        else:
            print self.style.NOTICE("\nWriting file:\n%s" % file)

            f = open(file, 'w')
            f.write(content_format % context)
            f.close()

    def __confirm_creation(self, dirs_to_create, files_to_create):
        """
        Make sure that the user wants to do this
        """
        cwd = os.getcwd()
        shorthand = "."

        answer = raw_input(self.style.NOTICE(
            MESSAGE_CONFIRM_CREATION % (
                "\n".join(dirs_to_create).replace(cwd, shorthand),
                "\n".join(files_to_create).replace(cwd, shorthand),
            )))

        if answer.lower.startswith('y'):
            return True

        return False

    def __create_these(self, dirs_to_create, files_to_create):
        """
        Makes the directories and files needed by Django Skylark
        """
        already_exists_message = "%s already exists, skiping"

        for dir in dirs_to_create:
            if os.path.isdir(dir):
                print self.style.ERROR(already_exists_message % dir)
                continue

            os.makedirs(dir)

        for file in files_to_create:
            if os.path.isfile(file):
                print self.style.ERROR(already_exists_message % file)
                continue

            open(file, 'a').close()



from django.conf.urls.defaults import *
from dummyapp.views import index

from skylark.views.generic import *
from skylark import chirp

chirp.autodiscover()

handler404 = render_404_from_yaml('dummyapp/handler/404.yaml')
handler500 = render_500_from_yaml('dummyapp/handler/500.yaml')

urlpatterns = patterns('',
    (r'^$', 'dummyapp.views.index'),
    (r'^badview$', 'dummyapp.views.badview'),
    (r'^planapp$', 'planapp.views.index'),
)



from setuptools import setup, find_packages
import os

from pkginfo import version

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README.rst')).read()
NEWS = open(os.path.join(here, 'NEWS.rst')).read()

install_requires = [
    # http://packages.python.org/distribute/setuptools.html#declaring-dependencies
    'pyyaml==3.09',
    'pytidylib==0.2.1',
]

setup(
    name='django-skylark',
    version=version,
    description="",
    long_description=README + '\n\n' + NEWS,
    classifiers=[], # Get strings from http://pypi.python.org/pypi?%3Aaction=list_classifiers
    keywords='web 2 html css mvc django dojo yaml',
    author='Rob Madole',
    author_email='robmadole@gmail.com',
    url='http://robmadole.com',
    license='BSD',
    packages=find_packages('src'),
    package_dir = {'': 'src'},
    include_package_data=True,
    zip_safe=False,
    install_requires=[
      'pyyaml==3.09',
      'pytidylib==0.2.1',
    ],
    entry_points={}
)
